/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff;

import java.net.URI;
import java.util.Collection;
import java.util.HashSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.model.OWLAntiSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLAxiomAnnotationAxiom;
import org.semanticweb.owl.model.OWLAxiomVisitor;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAssertionAxiom;
import org.semanticweb.owl.model.OWLConstant;
import org.semanticweb.owl.model.OWLDataAllRestriction;
import org.semanticweb.owl.model.OWLDataExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataProperty;
import org.semanticweb.owl.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLDataPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLDataPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLDataRange;
import org.semanticweb.owl.model.OWLDataSomeRestriction;
import org.semanticweb.owl.model.OWLDataSubPropertyAxiom;
import org.semanticweb.owl.model.OWLDataType;
import org.semanticweb.owl.model.OWLDataValueRestriction;
import org.semanticweb.owl.model.OWLDeclarationAxiom;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLDescriptionVisitor;
import org.semanticweb.owl.model.OWLDifferentIndividualsAxiom;
import org.semanticweb.owl.model.OWLDisjointClassesAxiom;
import org.semanticweb.owl.model.OWLDisjointDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointUnionAxiom;
import org.semanticweb.owl.model.OWLEntityAnnotationAxiom;
import org.semanticweb.owl.model.OWLEntityVisitor;
import org.semanticweb.owl.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owl.model.OWLEquivalentDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLEquivalentObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLFunctionalDataPropertyAxiom;
import org.semanticweb.owl.model.OWLFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLInverseFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLIrreflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLNegativeDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLNegativeObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectAllRestriction;
import org.semanticweb.owl.model.OWLObjectComplementOf;
import org.semanticweb.owl.model.OWLObjectExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectOneOf;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyChainSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyExpression;
import org.semanticweb.owl.model.OWLObjectPropertyInverse;
import org.semanticweb.owl.model.OWLObjectPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLObjectSelfRestriction;
import org.semanticweb.owl.model.OWLObjectSomeRestriction;
import org.semanticweb.owl.model.OWLObjectSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectUnionOf;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntologyAnnotationAxiom;
import org.semanticweb.owl.model.OWLPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLPropertyExpression;
import org.semanticweb.owl.model.OWLPropertyExpressionVisitor;
import org.semanticweb.owl.model.OWLReflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLSameIndividualsAxiom;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.model.OWLSubPropertyAxiom;
import org.semanticweb.owl.model.OWLSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTransitiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTypedConstant;
import org.semanticweb.owl.model.OWLUnaryPropertyAxiom;
import org.semanticweb.owl.model.OWLUntypedConstant;
import org.semanticweb.owl.model.SWRLRule;

public class DLSyntax implements Syntax {

	private static final Log LOG = LogFactory.getLog(DLSyntax.class);

	private static final String SUB = "\u2291";
	private static final String EQUIVALENT = "\u2261";
	private static final String NEGATION = "\u00ac";

	/**
	 * {@inheritDoc}
	 */
	public String writeAxiom(final OWLAxiom axiom, final boolean fullURI,
			final Object context, final boolean html) {
		if (axiom == null) {
			throw new IllegalArgumentException();
		}

		final StringBuilder b = new StringBuilder();

		axiom.accept(new OWLAxiomVisitor() {

			private void propertyCharacteristics(final String type,
					final OWLUnaryPropertyAxiom<?> p, final boolean fullURI) {
				if (context == null) {
					b.append(type).append("(").append(
							write(p.getProperty(), fullURI, html)).append(")");
				} else if (context.equals(p.getProperty())) {
					b.append(type);
				} else {
					LOG.warn("The context '" + context
							+ "' is not equal to the relation of " + p);
					b.append(p.toString());
				}
			}

			private void symmetricConceptAxiom(
					final Collection<OWLDescription> de, final Object context,
					final String separator, final boolean fullURI) {
				if (new HashSet<Object>(de).contains(context)) { // TODO remove
					// - for
					// cases
					// that de
					// is a
					// TreeSet
					// and
					// context
					// is not
					// Comparable
					for (final OWLDescription d : de) {
						if (!d.equals(context)) {
							b.append(separator).append(write(d, fullURI, html));
						}
					}
				} else {
					if (context != null) {
						LOG.warn("The context '" + context
								+ "' is not equal to an element of " + de);
					}
					boolean first = true;

					for (final OWLDescription d : de) {
						if (!first) {
							b.append(separator);
						}
						first = false;
						b.append(write(d, fullURI, html));
					}
				}
			}

			private void symmetricPropertiesAxiom(
					final Collection<? extends OWLPropertyExpression<?, ?>> de,
					final Object context, final String separator,
					final boolean fullURI) {
				if (new HashSet<Object>(de).contains(context)) { // TODO remove
					// - for
					// cases
					// that de
					// is a
					// TreeSet
					// and
					// context
					// is not
					// Comparable
					for (final OWLPropertyExpression<?, ?> d : de) {
						if (!d.equals(context)) {
							b.append(separator).append(write(d, fullURI, html));
						}
					}
				} else {
					if (context != null) {
						LOG.warn("The context '" + context
								+ "' is not equal to an element of " + de);
					}
					boolean first = true;

					for (final OWLPropertyExpression<?, ?> d : de) {
						if (!first) {
							b.append(separator);
						}
						first = false;
						b.append(write(d, fullURI, html));
					}
				}
			}

			private void symmetricIndividualsAxiom(
					final Collection<OWLIndividual> de, final Object context,
					final String separator, final boolean fullURI) {
				if (de.contains(context)) {
					for (final OWLIndividual d : de) {
						b.append(separator).append(write(d, fullURI, html));
					}
				} else {
					if (context != null) {
						LOG.warn("The context '" + context
								+ "' is not equal to an element of " + de);
					}
					boolean first = true;

					for (final OWLIndividual d : de) {
						if (!first) {
							b.append(separator);
						}
						first = false;
						b.append(write(d, fullURI, html));
					}
				}
			}

			private void domainAxiom(OWLPropertyDomainAxiom<?> arg0) {
				if (arg0.getProperty().equals(context)) {
					b.append("Domain=").append(
							write(arg0.getDomain(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'property' part of the axiom "
										+ arg0);
					}

					b.append("Domain(").append(
							write(arg0.getProperty(), fullURI, html)).append(
							")=")
							.append(write(arg0.getDomain(), fullURI, html));
				}
			}

			private void propertyValue(final boolean object,
					final boolean notNeg,
					final OWLPropertyAssertionAxiom<?, ?> arg0) {

				String o;

				if (object) {
					o = write(((OWLIndividual) arg0.getObject()), fullURI, html);
				} else {
					o = write(((OWLConstant) arg0.getObject()), fullURI, html);
				}

				if (arg0.getSubject().equals(context)) {
					if (!notNeg)
						b.append(NEGATION);

					b.append(write(arg0.getProperty(), fullURI, html)).append(
							" \u2192 ").append(o);
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'subejct' part of the axiom "
										+ arg0);
					}

					if (!notNeg)
						b.append(NEGATION);

					b.append(write(arg0.getProperty(), fullURI, html)).append(
							"(")
							.append(write(arg0.getSubject(), fullURI, html))
							.append(", ").append(o).append(")");
				}
			}

			private void subProperty(OWLSubPropertyAxiom<?> arg0) {
				if (arg0.getSubProperty().equals(context)) {
					b.append(" ").append(SUB).append(" ").append(
							write(arg0.getSuperProperty(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'sub' part of the axiom "
										+ arg0);
					}

					b.append(write(arg0.getSubProperty(), fullURI, html))
							.append(" ").append(SUB).append(" ").append(
									write(arg0.getSuperProperty(), fullURI,
											html));
				}
			}

			public void visit(OWLSubClassAxiom arg0) {
				if (arg0.getSubClass().equals(context)) {
					b.append(" ").append(SUB).append(" ").append(
							write(arg0.getSuperClass(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'sub' part of the axiom "
										+ arg0);
					}

					b.append(write(arg0.getSubClass(), fullURI, html)).append(
							" ").append(SUB).append(" ").append(
							write(arg0.getSuperClass(), fullURI, html));
				}
			}

			public void visit(OWLNegativeObjectPropertyAssertionAxiom arg0) {
				propertyValue(true, false, arg0);
			}

			public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
				propertyCharacteristics("anti-symmetric", arg0, fullURI);
			}

			public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
				propertyCharacteristics("reflexive", arg0, fullURI);
			}

			public void visit(OWLDisjointClassesAxiom arg0) {
				symmetricConceptAxiom(arg0.getDescriptions(), context,
						" \u22A0 ", fullURI);
			}

			public void visit(OWLDataPropertyDomainAxiom arg0) {
				domainAxiom(arg0);
			}

			public void visit(OWLOntologyAnnotationAxiom arg0) {
				b.append("Annotation").append(" ");
				b.append(" ").append(
						getName(arg0.getAnnotation().getAnnotationURI(),
								fullURI));

				if (arg0.getAnnotation().isAnnotationByConstant()) {
					b.append(" ").append(
							write(arg0.getAnnotation()
									.getAnnotationValueAsConstant(), fullURI,
									html));
				} else {
					// TODO
					b.append(" \"").append(
							arg0.getAnnotation().getAnnotationValue()).append(
							"\"");
				}
			}

			public void visit(OWLImportsDeclaration arg0) {
				b.append("imports").append(" ");
				b.append(arg0.getImportedOntologyURI());
				// TODO
			}

			public void visit(OWLAxiomAnnotationAxiom arg0) {
				b.append("[").append(
						getName(arg0.getAnnotation().getAnnotationURI(),
								fullURI)).append("]");

				b.append(" \"");
				if (arg0.getAnnotation().isAnnotationByConstant()) {
					b.append(write(arg0.getAnnotation()
							.getAnnotationValueAsConstant(), fullURI, html));
				} else {
					b.append(arg0.getAnnotation().getAnnotationValue());
				}
				b.append("\"");
			}

			public void visit(OWLObjectPropertyDomainAxiom arg0) {
				domainAxiom(arg0);
			}

			public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
				symmetricPropertiesAxiom(arg0.getProperties(), context, " "
						+ EQUIVALENT + " ", fullURI);
			}

			public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
				propertyValue(false, false, arg0);
			}

			public void visit(OWLDifferentIndividualsAxiom arg0) {
				symmetricIndividualsAxiom(arg0.getIndividuals(), context,
						" \u2260 ", fullURI);
			}

			public void visit(OWLDisjointDataPropertiesAxiom arg0) {
				symmetricPropertiesAxiom(arg0.getProperties(), context,
						" \u22A0 ", fullURI);
			}

			public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
				symmetricPropertiesAxiom(arg0.getProperties(), context,
						" \u22A0 ", fullURI);
			}

			public void visit(OWLObjectPropertyRangeAxiom arg0) {
				if (arg0.getProperty().equals(context)) {
					b.append("Range=").append(
							write(arg0.getRange(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'property' part of the axiom "
										+ arg0);
					}

					b.append("Range(").append(
							write(arg0.getProperty(), fullURI, html)).append(
							")=").append(write(arg0.getRange(), fullURI, html));
				}
			}

			public void visit(OWLObjectPropertyAssertionAxiom arg0) {
				propertyValue(true, true, arg0);
			}

			public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
				propertyCharacteristics("functional", arg0, fullURI);

			}

			public void visit(OWLObjectSubPropertyAxiom arg0) {
				subProperty(arg0);
			}

			public void visit(OWLDisjointUnionAxiom arg0) {
				b.append(write(arg0.getOWLClass(), fullURI, html)).append(
						"=DisjointUnionOf(");

				boolean first = true;

				for (final OWLDescription d : arg0.getDescriptions()) {
					if (!first) {
						b.append(", ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append(")");
				// TODO
			}

			public void visit(OWLDeclarationAxiom arg0) {
				b.append("declares ").append(" ");
				b.append(getName(arg0.getEntity().getURI(), fullURI));
			}

			public void visit(OWLEntityAnnotationAxiom arg0) {
				if (context == null) {
					arg0.getSubject().accept(new OWLEntityVisitor() {

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLClass cls) {
							// TODO Auto-generated method stub
							b.append(write(cls, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLObjectProperty property) {
							b.append(write(property, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLDataProperty property) {
							b.append(write(property, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLIndividual individual) {
							b.append(write(individual, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLDataType dataType) {
							b.append(write(dataType, fullURI, html));
						}
					});

					b.append(" ");
				}

				b.append("[").append(
						getName(arg0.getAnnotation().getAnnotationURI(),
								fullURI)).append("]");
				if (arg0.getAnnotation().isAnnotationByConstant()) {
					b.append(" ").append(
							write(arg0.getAnnotation()
									.getAnnotationValueAsConstant(), fullURI,
									html));
				} else {
					// TODO
					b.append(" ").append(
							arg0.getAnnotation().getAnnotationValue()).append(
							"\"");
				}
			}

			public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
				propertyCharacteristics("symmetric", arg0, fullURI);
			}

			public void visit(OWLDataPropertyRangeAxiom arg0) {
				if (arg0.getProperty().equals(context)) {
					b.append("Range=").append(
							write(arg0.getRange(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'property' part of the axiom "
										+ arg0);
					}

					b.append("Range(").append(
							write(arg0.getProperty(), fullURI, html)).append(
							")=").append(write(arg0.getRange(), fullURI, html));
				}

			}

			public void visit(OWLFunctionalDataPropertyAxiom arg0) {
				propertyCharacteristics("functional", arg0, fullURI);
			}

			public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
				symmetricPropertiesAxiom(arg0.getProperties(), context, " "
						+ EQUIVALENT + " ", fullURI);
			}

			public void visit(OWLClassAssertionAxiom arg0) {
				if (arg0.getIndividual().equals(context)) {
					b.append(" \u2208 ").append(
							write(arg0.getDescription(), fullURI, html));
				} else {
					if (context != null) {
						LOG
								.warn("The context '"
										+ context
										+ "' is not equal to the 'property' part of the axiom "
										+ arg0);
					}

					b.append(write(arg0.getDescription(), fullURI, html))
							.append("(").append(
									write(arg0.getIndividual(), fullURI, html))
							.append(")");
				}
			}

			public void visit(OWLEquivalentClassesAxiom arg0) {
				symmetricConceptAxiom(arg0.getDescriptions(), context, " "
						+ EQUIVALENT + " ", fullURI);
			}

			public void visit(OWLDataPropertyAssertionAxiom arg0) {
				propertyValue(false, true, arg0);
			}

			public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
				propertyCharacteristics("transitive", arg0, fullURI);
			}

			public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
				propertyCharacteristics("irreflexive", arg0, fullURI);
			}

			public void visit(OWLDataSubPropertyAxiom arg0) {
				subProperty(arg0);
			}

			public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
				propertyCharacteristics("inverse-functional", arg0, fullURI);
			}

			public void visit(OWLSameIndividualsAxiom arg0) {
				symmetricIndividualsAxiom(arg0.getIndividuals(), context,
						" = ", fullURI);
			}

			public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
				boolean first = true;
				for (final OWLObjectPropertyExpression e : arg0
						.getPropertyChain()) {
					if (!first) {
						b.append(" \u2218 ");
					}
					first = false;
					b.append(write(e, fullURI, html));
				}

				b.append(" \u2291 ").append(
						write(arg0.getSuperProperty(), fullURI, html));
				// TODO
			}

			public void visit(OWLInverseObjectPropertiesAxiom arg0) {
				symmetricPropertiesAxiom(arg0.getProperties(), context,
						" inverse-of ", fullURI);
			}

			public void visit(SWRLRule arg0) {
				b.append(arg0.toString());
				// TODO
			}
		});

		return b.toString();
	}

	private String write(final OWLDescription concept, final boolean fullURI,
			final boolean html) {
		if (concept == null) {
			throw new IllegalArgumentException();
		}

		final StringBuilder b = new StringBuilder();

		concept.accept(new OWLDescriptionVisitor() {

			public void visit(OWLClass arg0) {
				if (arg0.isOWLThing()) {
					b.append("\u22a4");
				} else if (arg0.isOWLNothing()) {
					b.append("\u22a5");
				} else {
					b.append(getName(arg0.getURI(), fullURI));
				}
			}

			public void visit(OWLObjectIntersectionOf arg0) {
				b.append("(");
				boolean first = true;

				for (final OWLDescription d : arg0.getOperands()) {
					if (!first) {
						b.append(" \u2293 ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append(")");
			}

			public void visit(OWLObjectUnionOf arg0) {
				b.append("(");
				boolean first = true;

				for (final OWLDescription d : arg0.getOperands()) {
					if (!first) {
						b.append(" \u2294 ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append(")");
			}

			public void visit(OWLObjectComplementOf arg0) {
				b.append(NEGATION).append(
						write(arg0.getOperand(), fullURI, html));
			}

			public void visit(OWLObjectSomeRestriction arg0) {
				b.append("(\u2203 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5")
						.append(write(arg0.getFiller(), fullURI, html)).append(
								")");
			}

			public void visit(OWLObjectAllRestriction arg0) {
				b.append("(\u2200 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5")
						.append(write(arg0.getFiller(), fullURI, html)).append(
								")");
			}

			public void visit(OWLObjectValueRestriction arg0) {
				b.append("(\u2203 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5{")
						.append(write(arg0.getValue(), fullURI, html)).append(
								"})");
			}

			public void visit(OWLObjectMinCardinalityRestriction arg0) {
				b.append("(\u2265 ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));
				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}

			public void visit(OWLObjectExactCardinalityRestriction arg0) {
				b.append("(= ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));
				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}

			public void visit(OWLObjectMaxCardinalityRestriction arg0) {
				b.append("(\u2264 ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));
				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}

			public void visit(OWLObjectSelfRestriction arg0) {
				b.append("(\u2203 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5").append("{SELF}").append(")");
			}

			public void visit(OWLObjectOneOf arg0) {
				b.append("{");
				boolean first = true;

				for (final OWLIndividual d : arg0.getIndividuals()) {
					if (!first) {
						b.append(", ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append("}");
			}

			public void visit(OWLDataSomeRestriction arg0) {
				b.append("(\u2203 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5")
						.append(write(arg0.getFiller(), fullURI, html)).append(
								")");
			}

			public void visit(OWLDataAllRestriction arg0) {
				b.append("(\u2200 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5")
						.append(write(arg0.getFiller(), fullURI, html)).append(
								")");
			}

			public void visit(OWLDataValueRestriction arg0) {
				b.append("(\u2203 ").append(
						write(arg0.getProperty(), fullURI, html)).append(
						"\u22c5{")
						.append(write(arg0.getValue(), fullURI, html)).append(
								"})");
			}

			public void visit(OWLDataMinCardinalityRestriction arg0) {
				b.append("(\u2265 ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));

				if (arg0.isQualified()) {
					b.append("\u22c5").append(
							write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}

			public void visit(OWLDataExactCardinalityRestriction arg0) {
				b.append("(= ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));

				if (arg0.isQualified()) {
					b.append("\u22c5").append(
							write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}

			public void visit(OWLDataMaxCardinalityRestriction arg0) {
				b.append("(\u2264 ").append(arg0.getCardinality()).append(" ")
						.append(write(arg0.getProperty(), fullURI, html));
				if (arg0.isQualified()) {
					b.append("\u22c5").append(
							write(arg0.getFiller(), fullURI, html));
				}
				b.append(")");
			}
		});

		return b.toString();
	}

	public String write(final OWLConstant c, final boolean fullURI,
			final boolean html) {
		if (c.isTyped()) {
			final OWLTypedConstant tc = c.asOWLTypedConstant();
			return "\"" + tc.getLiteral() + "\"^^"
					+ write(tc.getDataType(), fullURI, html);
		} else {
			final OWLUntypedConstant tc = c.asOWLUntypedConstant();
			return "\"" + tc.getLiteral() + "\"";
		}
	}

	private String write(final OWLDataRange dt, final boolean fullURI,
			final boolean html) {
		return dt.toString();
	}

	private String write(final OWLPropertyExpression<?, ?> relation,
			final boolean full, final boolean html) {
		if (relation == null) {
			throw new IllegalArgumentException("Relation is null.");
		}

		final StringBuffer b = new StringBuffer();

		relation.accept(new OWLPropertyExpressionVisitor() {

			public void visit(OWLObjectProperty arg0) {
				b.append(getName(arg0.getURI(), full));
			}

			public void visit(OWLObjectPropertyInverse arg0) {
				b.append(write(arg0.getInverse(), full, html)).append("\u02c9");
			}

			public void visit(OWLDataProperty arg0) {
				b.append(getName(arg0.getURI(), full));
			}

		});

		return b.toString();
	}

	private String write(OWLIndividual individual, boolean full,
			final boolean html) {
		return getName(individual.getURI(), full);
	}

	private static String getName(final URI uri, final boolean fullURI) {
		if (fullURI) {
			return uri.toString();
		} else if (uri.getFragment() == null) {
			return uri.getPath().substring(uri.getPath().lastIndexOf('/') + 1);
		} else {

			return uri.getFragment();
		}
	}
}
