/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextPane;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.semanticweb.owl.inference.OWLReasoner;
import org.semanticweb.owl.inference.OWLReasonerException;
import org.semanticweb.owl.inference.OWLReasonerFactory;
import org.semanticweb.owl.model.AddAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLException;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLNamedObject;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.RemoveAxiom;

import cz.cvut.kbss.owldiff.OWLAxiomTreeModel.NodeModel;
import cz.cvut.kbss.owldiff.pellet.Diff;
import cz.cvut.kbss.owldiff.pellet.IncOWLAPIExplanationManager;
import cz.cvut.kbss.owldiff.pellet.PelletReasonerFactory;

public class DiffView extends JPanel {

	private static final long serialVersionUID = -1511915926159556871L;

	private static final String MERGE_DIR = "ontologies/";

	private OWLOntology updateO;
	private OWLOntologyManager updateM;

	private OWLOntology originalO;
	private OWLOntologyManager originalM;

	private Set<OWLAxiom> axiomsToAdd, axiomsToDelete;

	private Diff d;
	private cz.cvut.kbss.owldiff.cex.Diff cexDiffOriginal;
	private cz.cvut.kbss.owldiff.cex.Diff cexDiffUpdate;
	private final Map<OWLAxiom, Set<OWLAxiom>> explanations = new HashMap<OWLAxiom, Set<OWLAxiom>>();

	private URI oldFile, newFile;

	private final OWLReasonerFactory f = new PelletReasonerFactory();

	// VIEW

	private JTree treOriginal = null;
	private JScrollPane scpOriginal = null;
	private OWLAxiomTreeModel tmOriginal = null;

	private JTree treUpdate = null;
	private JScrollPane scpUpdate = null;
	private OWLAxiomTreeModel tmUpdate = null;

	private JTextPane originalInfo;
	private JTextPane updateInfo;

	private JLabel originalTitleLabel, updateTitleLabel;

	// internal view state
	private DiffVisualization view = DiffVisualization.SIMPLE_FRAME_VIEW;
	private SyntaxEnum syntax;
	private boolean synchronizing = false;
	private DiffView thisPanel = this;
	private boolean showCommonAxioms;
	private final Framework fr;

	public DiffView(final Framework f) {
		super();

		this.fr = f;

		createView();

		axiomsToAdd = new HashSet<OWLAxiom>();
		axiomsToDelete = new HashSet<OWLAxiom>();

		syntax = SyntaxEnum.MANCHESTER;
	}

	private void changeView(final DiffVisualization v) {
		if (thisPanel.view == v) {
			return;
		}

		thisPanel.view = v;
		new Thread() {
			public void run() {
				fr.setProgressVisible(true);
				fr.setProgress(0);
				fr.setProgressMax(2);
				fr.showMsg("Building view ...");
				tmOriginal.setView(view);
				fr.setProgress(1);
				tmUpdate.setView(view);
				fr.setProgress(2);
				// getOriginal().expandRow(0);
				// getUpdate().expandRow(0);
				fr.showMsg("View created.");
				fr.setProgressVisible(false);
			}
		}.start();
	}

	public void updateDependencies() {
		fr.setEnabled(OWLDiffAction.merge, !axiomsToAdd.isEmpty()
				|| !axiomsToDelete.isEmpty());

		fr.setEnabled(OWLDiffAction.useCEX, cexDiffOriginal == null
				&& tmUpdate != null);
		fr.setEnabled(OWLDiffAction.showExplanations, d != null
				&& !d.isEntailmentComputed());
		fr.setEnabled(OWLDiffAction.descriptionLogic, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.manchester, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.showAxiomList, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.showAssertedFrames, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.showClassifiedFrames, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.showCommon, tmUpdate != null);
		fr.setEnabled(OWLDiffAction.selectAllOriginal, d != null
				&& !d.getOriginalDiff().isEmpty());
		fr.setEnabled(OWLDiffAction.selectAllUpdate, d != null
				&& !d.getUpdateDiff().isEmpty());
		fr
				.setEnabled(OWLDiffAction.deselectAllOriginal, !axiomsToAdd
						.isEmpty());
		fr.setEnabled(OWLDiffAction.deselectAllUpdate, !axiomsToDelete
				.isEmpty());
	}

	private void createView() {
		this.setLayout(new BorderLayout());
		final JPanel pnlTop = new JPanel();
		pnlTop.setLayout(new BorderLayout());
		final JPanel pnlButtons = new JPanel();
		pnlButtons.setLayout(new BorderLayout());
		pnlTop.add(pnlButtons, BorderLayout.CENTER);
		final JPanel pnlLeftTop = new JPanel();
		pnlLeftTop.setLayout(new FlowLayout());
		originalTitleLabel = new JLabel("original");
		pnlLeftTop.add(originalTitleLabel);
		pnlTop.add(pnlLeftTop, BorderLayout.WEST);
		final JPanel pnlRightTop = new JPanel();
		pnlRightTop.setLayout(new FlowLayout());
		updateTitleLabel = new JLabel("update");
		pnlRightTop.add(updateTitleLabel);
		pnlTop.add(pnlRightTop, BorderLayout.EAST);
		this.add(pnlTop, BorderLayout.NORTH);
		this.add(createDiffSplitPane(), BorderLayout.CENTER);
	}

	private JSplitPane createDiffSplitPane() {
		final JSplitPane jSplitPane = new JSplitPane();
		jSplitPane.setOneTouchExpandable(true);
		jSplitPane.setDividerSize(10);

		scpOriginal = new JScrollPane(getOriginal());
		scpOriginal.getVerticalScrollBar()
				.addAdjustmentListener(new Listener());
		scpUpdate = new JScrollPane(getUpdate());
		scpUpdate.getVerticalScrollBar().addAdjustmentListener(new Listener());

		originalInfo = new JTextPane();
		originalInfo.setEditable(false);
		originalInfo.setEditorKit(new HTMLEditorKit());
		originalInfo.setVisible(false);

		final JPanel jLeftPane = new JPanel();
		jLeftPane.setLayout(new BorderLayout());
		jLeftPane.add(scpOriginal, BorderLayout.CENTER);
		jLeftPane.add(originalInfo, BorderLayout.SOUTH);

		updateInfo = new JTextPane();
		updateInfo.setEditable(false);
		updateInfo.setEditorKit(new HTMLEditorKit());
		updateInfo.setVisible(false);

		final JPanel jRightPane = new JPanel();
		jRightPane.setLayout(new BorderLayout());
		jRightPane.add(scpUpdate, BorderLayout.CENTER);
		jRightPane.add(updateInfo, BorderLayout.SOUTH);

		jSplitPane.setLeftComponent(jLeftPane);
		jSplitPane.setRightComponent(jRightPane);

		jSplitPane.setResizeWeight(0.5D);
		return jSplitPane;
	}

	private TreePath getTreePath(JTree tree, TreePath path) {
		Object[] o = new Object[path.getPathCount()];
		o[0] = tree.getModel().getRoot();

		for (int i = 0; i < path.getPathCount() - 1; i++) {

			boolean found = false;

			for (int j = 0; j < tree.getModel().getChildCount(o[i]); j++) {
				if (tree.getModel().getChild(o[i], j).equals(
						path.getPathComponent(i + 1))) {
					o[i + 1] = tree.getModel().getChild(o[i], j);
					found = true;
					break;
				}
			}
			if (!found) {
				return null;
			}
		}
		return new TreePath(o);
	}

	private void expandTree(JTree tree, TreePath path) {
		if (!synchronizing) {
			synchronizing = true;
		} else {
			return;
		}
		path = getTreePath(tree, path);
		if (path != null) {
			tree.expandPath(path);
		}
		synchronizing = false;
	}

	private void collapseTree(JTree tree, TreePath path) {
		if (!synchronizing) {
			synchronizing = true;
		} else {
			return;
		}
		path = getTreePath(tree, path);
		if (path != null) {
			tree.collapsePath(path);
		}
		synchronizing = false;
	}

	private void selectItem(JTree tree, TreePath path) {
		if (!synchronizing) {
			synchronizing = true;
		} else {
			return;
		}
		path = getTreePath(tree, path);
		if (path != null) {
			tree.setSelectionPath(path);
		} else {
			tree.clearSelection();
			originalInfo.setText(null);
			updateInfo.setText(null);
		}
		synchronizing = false;
	}

	private void writeAxiomProperty(TreePath path, OWLAxiomTreeModel t,
			final JTextPane text, final OWLOntologyManager m,
			final OWLOntology o, final cz.cvut.kbss.owldiff.cex.Diff diff) {
		if (path.getLastPathComponent() instanceof OWLAxiom) {
			text.setVisible(true);
			final OWLAxiomTreeModel tm = (OWLAxiomTreeModel) t;
			final OWLAxiom ax = (OWLAxiom) path.getLastPathComponent();
			if (tm.isInferred(ax)) {
				if (!explanations.containsKey(ax)) {
					new Thread() {
						public void run() {
							explanations.put(ax,
									getExplanationManager(m, o, ax)
											.getNextExplanation());
							writeExplanation(text, explanations.get(ax));
						}
					}.start();
				} else {
					writeExplanation(text, explanations.get(ax));
				}
			} else if (tm.isCommon(ax)) {
				text.setVisible(false);
			} else {
				text
						.setText("This axiom has no connection to the other ontology");
			}
		} else if ((diff != null)
				&& (path.getLastPathComponent() instanceof NodeModel)
				&& (((NodeModel) path.getLastPathComponent()).getObject() instanceof OWLClass)) {
			text.setVisible(true);
			final OWLClass ax = (OWLClass) ((NodeModel) path
					.getLastPathComponent()).getObject();

			if (diff.getDiffR() != null && diff.getDiffR().contains(ax)) {
				if (diff.getDiffL() != null && diff.getDiffL().contains(ax)) {
					text
							.setText("Extension both contains and lacks some domain elements comparing to the other ontology.");
				} else {
					text
							.setText("Extension contains some extra domain elements comparing to the other ontology.");
				}
			} else if (diff.getDiffL() != null && diff.getDiffL().contains(ax)) {
				text
						.setText("Extension lacks some domain elements comparing to the other ontology.");
			} else {
				text.setVisible(false);
			}
		} else {
			text.setVisible(false);
		}
	}

	private void writeExplanation(final JTextPane text,
			final Set<OWLAxiom> axioms) {
		final StringBuffer b = new StringBuffer();
		b
				.append("<html><i>Inferred from the following axioms in the other ontology:</i><br/>");
		b.append("<ul>");

		for (final OWLAxiom a : axioms) {
			b.append("<li>");
			b.append(syntax.getSyntax().writeAxiom(a, false, null, true));
			b.append("</li>");
		}

		b.append("</ul></html>");

		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				text.setText(b.toString());
			}
		});
	}

	private ExplanationManager getExplanationManager(
			final OWLOntologyManager m, final OWLOntology o, final OWLAxiom a) {
		return new IncOWLAPIExplanationManager(m, o, a);
		// return new BBOWLAPIExplanationManager(m, o, a);
		// return new BBPelletExplanationManager(m, o, f, a);
	}

	private void changeViewInSwingThread(final DiffVisualization view) {
		final OWLDiffAction a;
		switch (view) {
		case LIST_VIEW:
			a = OWLDiffAction.showAxiomList;
			break;
		case SIMPLE_FRAME_VIEW:
			a = OWLDiffAction.showAssertedFrames;
			break;
		case CLASSIFIED_FRAME_VIEW:
			a = OWLDiffAction.showClassifiedFrames;
			break;
		default:
			a = null;
		}

		if (SwingUtilities.isEventDispatchThread()) {
			fr.executeAction(a);
		} else {
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					fr.executeAction(a);
				}
			});
		}
	}

	public void openFiles(final URI oldFile, final URI newFile) {
		this.oldFile = oldFile;
		this.newFile = newFile;

		if (view.equals(DiffVisualization.CLASSIFIED_FRAME_VIEW)) {
			changeViewInSwingThread(DiffVisualization.SIMPLE_FRAME_VIEW);
		}

		new Thread() {

			public void run() {
				try {
					parse();
					executeDiff();
				} catch (OWLDiffException e) {
					fr.showError(e, e.getMessage());
				}
			}
		}.start();
	}

	private void parse() throws OWLDiffException {
		fr.showMsg("Parsing ...");
		fr.setProgressVisible(true);
		fr.setProgressMax(4);
		fr.setProgress(0);

		try {
			originalM = org.semanticweb.owl.apibinding.OWLManager
					.createOWLOntologyManager();
			originalM.setSilentMissingImportsHandling(true);

			updateM = org.semanticweb.owl.apibinding.OWLManager
					.createOWLOntologyManager();
			updateM.setSilentMissingImportsHandling(true);

			originalO = originalM.loadOntology(oldFile);
			fr.showMsg("Original file succesfully parsed.");

			fr.setProgress(1);
			updateO = updateM.loadOntology(newFile);
			fr.showMsg("Update file succesfully parsed.");

			fr.setProgress(2);
		} catch (OWLException e) {
			fr.showError(e, "Error occured during loading");
			try {
				updateO = updateM.createOntology(URI.create(""));
				originalO = updateM.createOntology(URI.create(""));
			} catch (OWLOntologyCreationException e1) {
				fr.showError(e, e.getMessage());
			}
		} finally {
			fr.setProgress(4);
			fr.setProgressVisible(false);
		}

		// if (tryToResolveImports(originalM, originalO)) {
		// showMsg("Imports for original ontology succesfully resolved.");
		// } else {
		// showError(null,
		// "Error: imports for original ontology cannot be resolved.");
		// }
		// setProgress(3);
		//
		// if (tryToResolveImports(updateM, updateO)) {
		// showMsg("Imports for original ontology succesfully resolved.");
		// } else {
		// showError(null,
		// "Error: imports for update ontology cannot be resolved.");
		// }
	}

	private boolean tryToResolveImports(final OWLOntologyManager m,
			final OWLOntology o) {
		try {
			for (final OWLImportsDeclaration d : originalO
					.getImportsDeclarations()) {
				m.makeLoadImportRequest(d);
			}
			return true;
		} catch (Exception e) {
			fr.showError(e, e.getMessage());
			return false;
		}
	}

	/**
	 * This method initializes original
	 * 
	 * @return javax.swing.JList
	 */
	private JTree getOriginal() {
		if (treOriginal == null) {
			treOriginal = new JTree((TreeModel) null);
			treOriginal.setCellRenderer(new MyTreeCellRenderer());
			treOriginal.getSelectionModel().setSelectionMode(
					TreeSelectionModel.SINGLE_TREE_SELECTION);
			treOriginal.addTreeSelectionListener(new TreeSelectionListener() {

				public void valueChanged(TreeSelectionEvent arg0) {
					selectItem(treUpdate, arg0.getPath());
					writeAxiomProperty(arg0.getPath(), tmOriginal,
							originalInfo, updateM, updateO, cexDiffOriginal);
					updateInfo.setVisible(false);
				}
			});
			treOriginal.addTreeExpansionListener(new TreeExpansionListener() {

				public void treeExpanded(TreeExpansionEvent arg0) {
					expandTree(treUpdate, arg0.getPath());
				}

				public void treeCollapsed(TreeExpansionEvent arg0) {
					collapseTree(treUpdate, arg0.getPath());
				}
			});

			treOriginal.addMouseListener(new MouseAdapter() {
				/**
				 * {@inheritDoc}
				 */
				public void mousePressed(MouseEvent ev) {
					if (ev.getClickCount() == 2) {
						TreePath path = treOriginal.getPathForLocation(ev
								.getX(), ev.getY());
						if (path != null) {
							Object o = path.getLastPathComponent();
							if (o instanceof OWLAxiom
									&& d.getOriginalDiff().contains(o)) {
								OWLAxiom ax = (OWLAxiom) o;
								if (axiomsToAdd.contains(ax)) {
									axiomsToAdd.remove(ax);
								} else {
									axiomsToAdd.add(ax);
								}
							}

							updateDependencies();
						}
					}
				}
			});

			ToolTipManager.sharedInstance().registerComponent(treOriginal);
		}

		return treOriginal;
	}

	/**
	 * This method initializes update
	 * 
	 * @return javax.swing.JList
	 */
	private JTree getUpdate() {
		if (treUpdate == null) {
			treUpdate = new JTree((TreeModel) null);
			treUpdate.setCellRenderer(new MyTreeCellRenderer());
			treUpdate.getSelectionModel().setSelectionMode(
					TreeSelectionModel.SINGLE_TREE_SELECTION);
			treUpdate.addTreeSelectionListener(new TreeSelectionListener() {

				public void valueChanged(TreeSelectionEvent arg0) {
					selectItem(treOriginal, arg0.getPath());
					writeAxiomProperty(arg0.getPath(), tmUpdate, updateInfo,
							originalM, originalO, cexDiffUpdate);
					originalInfo.setVisible(false);
				}
			});
			treUpdate.addTreeExpansionListener(new TreeExpansionListener() {

				public void treeExpanded(TreeExpansionEvent arg0) {
					expandTree(treOriginal, arg0.getPath());
				}

				public void treeCollapsed(TreeExpansionEvent arg0) {
					collapseTree(treOriginal, arg0.getPath());
				}
			});

			treUpdate.addMouseListener(new MouseAdapter() {
				/**
				 * {@inheritDoc}
				 */
				public void mousePressed(MouseEvent ev) {
					if (ev.getClickCount() == 2) {
						TreePath path = treUpdate.getPathForLocation(ev.getX(),
								ev.getY());
						if (path != null) {
							Object o = path.getLastPathComponent();
							if (o instanceof OWLAxiom
									&& d.getUpdateDiff().contains(o)) {
								OWLAxiom ax = (OWLAxiom) o;
								if (axiomsToDelete.contains(ax)) {
									axiomsToDelete.remove(ax);
								} else {
									axiomsToDelete.add(ax);
								}
							}

							updateDependencies();
						}
					}
				}
			});

			ToolTipManager.sharedInstance().registerComponent(treUpdate);
		}
		return treUpdate;
	}

	private void executeCEXDiff() {
		new Thread() {

			public void run() {
				fr.setProgressVisible(true);
				fr.setProgress(0);
				fr.showMsg("Executing CEX diff ...");
				try {
					long time = System.currentTimeMillis();
					if (cexDiffOriginal == null) {
						cexDiffOriginal = new cz.cvut.kbss.owldiff.cex.Diff(fr);
					}
					cexDiffOriginal
							.diff(originalM, originalO, updateM, updateO);

					if (cexDiffUpdate == null) {
						cexDiffUpdate = new cz.cvut.kbss.owldiff.cex.Diff(fr);
					}

					cexDiffUpdate.diff(updateM, updateO, originalM, originalO);

					// if (!aShowCommon.getState()) {
					// aShowCommon.setState(true);
					// }

					if (view.equals(DiffVisualization.LIST_VIEW)) {
						changeViewInSwingThread(DiffVisualization.SIMPLE_FRAME_VIEW);
					}

					fr.showMsg("CEX diff took "
							+ (System.currentTimeMillis() - time) + " ms.");

					tmOriginal.fireTreeNodesChanged();
					tmUpdate.fireTreeNodesChanged();
				} catch (Exception ex) {
					fr.showError(ex, "CEX Failed");
				}
				fr.setProgressVisible(false);
			}
		}.start();
	}

	private void showEntailments() {
		new Thread() {

			public void run() {
				fr.setProgressVisible(true);
				fr.setProgress(0);
				fr.showMsg("Computing entailments ...");

				long time = System.currentTimeMillis();

				try {
					if (d == null) {
						d = new Diff(fr);
					}

					d.diff(originalM, originalO, updateM, updateO, true);

					tmOriginal.setInferred(d.getOriginalInferred());
					tmUpdate.setInferred(d.getUpdateInferred());

					fr.showMsg("Explanation generation took "
							+ (System.currentTimeMillis() - time) + " ms.");
				} catch (final Exception ex) {
					fr.showError(ex, ex.toString());
				}
				SwingUtilities.invokeLater(new Runnable() {

					public void run() {
						updateDependencies();
					}
				});
				fr.setProgressVisible(false);
			}
		}.start();
	}

	private void executeDiff() {
		fr.setProgressVisible(true);
		fr.setProgress(0);

		long time = System.currentTimeMillis();

		try {
			fr.showMsg("Executing diff ...");

			if (d == null) {
				d = new Diff(fr);
			}

			d.diff(originalM, originalO, updateM, updateO, false);

			fr.showMsg("Syntactic diff took "
					+ (System.currentTimeMillis() - time) + " ms.");

			axiomsToAdd.clear();
			axiomsToDelete.clear();

			tmOriginal = new OWLAxiomTreeModel(view, d.getOriginalDiff(),
					originalM, originalO, f);

			tmOriginal.setShowCommon(showCommonAxioms);

			tmUpdate = new OWLAxiomTreeModel(view, d.getUpdateDiff(), updateM,
					updateO, f);

			tmUpdate.setShowCommon(showCommonAxioms);

			getOriginal().setModel(tmOriginal);
			getUpdate().setModel(tmUpdate);

			cexDiffOriginal = null;
			cexDiffUpdate = null;

			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					originalTitleLabel.setText("original: " + oldFile);
					updateTitleLabel.setText("update: " + newFile);
				}
			});
		} catch (final Exception ex) {
			fr.showError(ex, ex.toString());
		}

		fr.setProgressVisible(false);

		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				updateInfo.setVisible(false);
				originalInfo.setVisible(false);
				updateDependencies();
			}
		});
	}

	private void saveResult() {
		new Thread() {

			public void run() {
				if (!"file".equals(newFile.getScheme())) {
					newFile = new File(MERGE_DIR + newFile.getPath()).toURI();
				}

				MergeAcceptDialog ma = new MergeAcceptDialog(
						syntax.getSyntax(), null, true, axiomsToAdd,
						axiomsToDelete, new File(newFile));
				if (ma.returnValue) {
					if (updateO == null) {
						return;
					}
					newFile = ma.getSelectedFile().toURI();
					if (axiomsToDelete != null) {
						for (OWLAxiom ax : axiomsToDelete) {
							try {
								updateM
										.applyChange(new RemoveAxiom(updateO,
												ax));
							} catch (Exception ex) {
								fr.showError(ex,
										"Failed to remove an axiom while saving: "
												+ ax);
								return;
							}
						}
					}
					if (axiomsToAdd != null) {
						for (OWLAxiom ax : axiomsToAdd) {
							try {
								updateM.applyChange(new AddAxiom(updateO, ax));
							} catch (Exception ex) {
								fr.showError(ex,
										"Failed to add an axiom while saving: "
												+ ax);
								return;
							}
						}
					}

					boolean merge = true;

					if (!isUpdateOntologyConsistent()) {
						if (JOptionPane
								.showConfirmDialog(thisPanel,
										"The resulting ontology is inconsistent, do you really want to save it.") != JOptionPane.OK_OPTION) {
							merge = false;
						}
					}

					if (merge) {
						try {
							fr.showMsg("Saving ... " + newFile.toString());
							updateM.saveOntology(updateO, newFile);
						} catch (Exception ex) {
							fr.showError(ex, "Failed to save merged ontology.");
							return;
						}
						fr.showMsg("Merged ontology was saved to '" + newFile
								+ "'");

						executeDiff();
					} else {
						revertUpdateOntology();
					}
				}
			}
		}.start();
	}

	private void revertUpdateOntology() {
		if (axiomsToDelete != null) {
			for (OWLAxiom ax : axiomsToDelete) {
				try {
					updateM.applyChange(new AddAxiom(updateO, ax));
				} catch (Exception ex) {
					fr
							.showError(ex,
									"Failed to restore an axiom while reverting: "
											+ ax);
					return;
				}
			}
		}
		if (axiomsToAdd != null) {
			for (OWLAxiom ax : axiomsToAdd) {
				try {
					updateM.applyChange(new RemoveAxiom(updateO, ax));
				} catch (Exception ex) {
					fr
							.showError(ex,
									"Failed to restore an axiom while reverting: "
											+ ax);
					return;
				}
			}
		}
	}

	private boolean isUpdateOntologyConsistent() {
		if (updateO == null) {
			return false;
		}
		try {
			final OWLReasoner r = f.createReasoner(updateM);
			r.loadOntologies(Collections.singleton(updateO));
			return r.isConsistent(updateO);
		} catch (OWLReasonerException e) {
			fr.showError(e, "Error when checking consistency.");
			return false;
		}
	}

	private class MyTreeCellRenderer extends DefaultTreeCellRenderer {

		private static final long serialVersionUID = -8026145488722024237L;

		public Component getTreeCellRendererComponent(JTree tree, Object value,
				boolean selected, boolean expanded, boolean leaf, int row,
				boolean hasFocus) {
			final TreeModel t = tree.getModel();

			setBackgroundNonSelectionColor(null);
			setTextNonSelectionColor(null);
			setTextSelectionColor(null);
			setBorderSelectionColor(null);

			String s = null;

			if (t instanceof OWLAxiomTreeModel) {
				final OWLAxiomTreeModel tm = (OWLAxiomTreeModel) t;

				if (value instanceof OWLAxiom) {
					final OWLAxiom ax = (OWLAxiom) value;
					final StringBuffer b = new StringBuffer();
					b.append("<html>");

					Object parent = null;

					if (!DiffVisualization.LIST_VIEW.equals(view)) {
						final TreePath tp = tree.getPathForRow(row);
						if (tp != null) {
							parent = ((NodeModel<?>) (tp.getPathComponent(tp
									.getPathCount() - 2))).getObject();
						}
					}

					b.append(syntax.getSyntax().writeAxiom(ax, false, parent,
							true));
					b.append("</html>");

					s = b.toString();

					if (tm.isCommon(ax)) {
						setTextSelectionColor(Color.BLUE);
						setTextNonSelectionColor(Color.BLUE);
					} else if (tm.isInferred(ax)) {
						setTextSelectionColor(Color.RED);
						setTextNonSelectionColor(Color.RED);
					} else {
						setTextSelectionColor(Color.GREEN);
						setTextNonSelectionColor(Color.GREEN);
					}
					if (axiomsToAdd.contains(value)) {
						setBackgroundNonSelectionColor(new Color(200, 255, 200));
						setBorderSelectionColor(Color.GREEN);
					} else if (axiomsToDelete.contains(value)) {
						setBackgroundNonSelectionColor(new Color(255, 200, 200));
						setBorderSelectionColor(Color.RED);
					}

					setToolTipText("<html>"
							+ syntax.getSyntax().writeAxiom(ax, true, null,
									true) + "</html>");

				} else {
					final NodeModel<?> m = (NodeModel<?>) value;
					final Object o = m.getObject();

					if (o instanceof OWLOntology) {
						s = ((OWLNamedObject) o).getURI().toString();
					} else if (o instanceof OWLNamedObject) {
						URI u = ((OWLNamedObject) o).getURI();
						if (u.getFragment() == null) {
							s = u.toString();
						} else {
							s = u.getFragment().toString();
						}
					} else {
						s = o.toString();
					}

					if (!t.getRoot().equals(value)) {
						if (m.containsInferred()) {
							setTextSelectionColor(Color.RED);
							setTextNonSelectionColor(Color.RED);
						} else if (m.containsDifferent()) {
							setTextSelectionColor(Color.GREEN);
							setTextNonSelectionColor(Color.GREEN);
						} else {
							setTextSelectionColor(Color.BLUE);
							setTextNonSelectionColor(Color.BLUE);
						}

						if (o instanceof OWLEntity) {
							setToolTipText(((OWLEntity) o).getURI().toString());
						}
					}
					if (o instanceof OWLClass) {
						cz.cvut.kbss.owldiff.cex.Diff cexDiff;

						if (tm == tmOriginal) {
							cexDiff = cexDiffOriginal;
						} else {
							cexDiff = cexDiffUpdate;
						}

						if (cexDiff != null && cexDiff.getDiffR() != null
								&& cexDiff.getDiffL() != null) {

							Color col = null;
							if (cexDiff.getDiffR().contains(o)) {
								if (cexDiff.getDiffL().contains(o)) {
									col = Color.GRAY;
								} else {
									col = Color.ORANGE;
								}
							} else if (cexDiff.getDiffL().contains(o)) {
								col = Color.YELLOW;
							}

							if (col != null) {
								setBackgroundNonSelectionColor(col);
								setBorderSelectionColor(col);
							}
						}
					}
				}
			}
			super.getTreeCellRendererComponent(tree, value, selected, expanded,
					leaf, row, hasFocus);
			setIcon(null);
			if (s != null) {
				setText(s);
			}

			return this;
		}
	}

	private class Listener implements AdjustmentListener {

		public void adjustmentValueChanged(AdjustmentEvent e) {
			scpOriginal.getVerticalScrollBar().setValue(e.getValue());
			scpUpdate.getVerticalScrollBar().setValue(e.getValue());
		}
	}

	public void deselectAllOriginalAxioms() {
		axiomsToAdd.removeAll(d.getOriginalDiff());
		updateDependencies();
		tmOriginal.fireTreeNodesChanged();
	}

	public void deselectAllUpdateAxioms() {
		axiomsToDelete.removeAll(d.getUpdateDiff());
		updateDependencies();
		tmUpdate.fireTreeNodesChanged();
	}

	public void runCEXDiff() {
		executeCEXDiff();
	}

	public void saveMergeResult() {
		saveResult();
	}

	public void selectAllOriginalAxioms() {
		axiomsToAdd.addAll(d.getOriginalDiff());
		updateDependencies();
		tmOriginal.fireTreeNodesChanged();
	}

	public void selectAllUpdateAxioms() {
		axiomsToDelete.addAll(d.getUpdateDiff());
		updateDependencies();
		tmUpdate.fireTreeNodesChanged();
	}

	public void setShowCommonAxioms(boolean show) {
		showCommonAxioms = show;

		if (tmOriginal != null) {
			tmOriginal.setShowCommon(show);
		}
		if (tmUpdate != null) {
			tmUpdate.setShowCommon(show);
		}
	}

	public boolean isShowCommonAxioms() {
		return showCommonAxioms;
	}

	public void setShowEntailments(boolean show) {
		showEntailments();
		updateDependencies();
	}

	public void setSyntax(SyntaxEnum s) {
		this.syntax = s;
		tmOriginal.fireTreeNodesChanged();
		tmUpdate.fireTreeNodesChanged();
	}

	public void setView(DiffVisualization v) {
		changeView(v);
	}

	public DiffVisualization getView() {
		return view;
	}

	public SyntaxEnum getSyntax() {
		return syntax;
	}
}
