/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff;

import java.net.URI;

import org.semanticweb.owl.model.OWLAntiSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLAxiomAnnotationAxiom;
import org.semanticweb.owl.model.OWLAxiomVisitor;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAssertionAxiom;
import org.semanticweb.owl.model.OWLConstant;
import org.semanticweb.owl.model.OWLDataAllRestriction;
import org.semanticweb.owl.model.OWLDataExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataProperty;
import org.semanticweb.owl.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLDataPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLDataPropertyExpression;
import org.semanticweb.owl.model.OWLDataPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLDataRange;
import org.semanticweb.owl.model.OWLDataSomeRestriction;
import org.semanticweb.owl.model.OWLDataSubPropertyAxiom;
import org.semanticweb.owl.model.OWLDataType;
import org.semanticweb.owl.model.OWLDataValueRestriction;
import org.semanticweb.owl.model.OWLDeclarationAxiom;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLDescriptionVisitor;
import org.semanticweb.owl.model.OWLDifferentIndividualsAxiom;
import org.semanticweb.owl.model.OWLDisjointClassesAxiom;
import org.semanticweb.owl.model.OWLDisjointDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointUnionAxiom;
import org.semanticweb.owl.model.OWLEntityAnnotationAxiom;
import org.semanticweb.owl.model.OWLEntityVisitor;
import org.semanticweb.owl.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owl.model.OWLEquivalentDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLEquivalentObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLFunctionalDataPropertyAxiom;
import org.semanticweb.owl.model.OWLFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLInverseFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLIrreflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLNegativeDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLNegativeObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectAllRestriction;
import org.semanticweb.owl.model.OWLObjectComplementOf;
import org.semanticweb.owl.model.OWLObjectExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectOneOf;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyChainSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyExpression;
import org.semanticweb.owl.model.OWLObjectPropertyInverse;
import org.semanticweb.owl.model.OWLObjectPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLObjectSelfRestriction;
import org.semanticweb.owl.model.OWLObjectSomeRestriction;
import org.semanticweb.owl.model.OWLObjectSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectUnionOf;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntologyAnnotationAxiom;
import org.semanticweb.owl.model.OWLPropertyExpression;
import org.semanticweb.owl.model.OWLPropertyExpressionVisitor;
import org.semanticweb.owl.model.OWLReflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLSameIndividualsAxiom;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.model.OWLSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTransitiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTypedConstant;
import org.semanticweb.owl.model.OWLUntypedConstant;
import org.semanticweb.owl.model.SWRLRule;

/**
 * 
 * @author Chucky
 */
public class ManchesterSyntax implements Syntax {

	private String keyword(final String s, final boolean html) {
		if (html) {
			return "<b>" + s + "</b>";
		}
		return s;
	}

	/**
	 * {@inheritDoc}
	 */
	public String writeAxiom(OWLAxiom axiom, boolean fullURI, Object context,
			final boolean html) {
		return _writeAxiom(axiom, fullURI, context, html);
	}

	private String _writeAxiom(final OWLAxiom axiom, final boolean fullURI,
			final Object context, final boolean html) {
		if (axiom == null) {
			throw new IllegalArgumentException();
		}

		final StringBuilder b = new StringBuilder();

		axiom.accept(new OWLAxiomVisitor() {
			public void visit(OWLSubClassAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getSubClass(), fullURI, html)).append(
							" ");
				}

				b.append(keyword("SubClassOf", html)).append(" ").append(
						write(arg0.getSuperClass(), fullURI, html));

			}

			public void visit(OWLDisjointClassesAxiom arg0) {
				b.append(keyword("DisjointWith", html)).append(" ");
				boolean first = true;
				for (final OWLDescription d : arg0.getDescriptions()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI, html));
						first = false;
					}
				}
			}

			public void visit(OWLEquivalentClassesAxiom arg0) {
				b.append(keyword("EquivalentTo", html)).append(" ");
				boolean first = true;
				for (final OWLDescription d : arg0.getDescriptions()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI, html));
						first = false;
					}
				}
			}

			public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html))
						.append(" Functional");
			}

			public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html)).append(" Symmetric");
			}

			public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html))
						.append(" Transitive");
			}

			public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html)).append(
						" Irreflexive");
			}

			public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html)).append(
						" AntiSymmetric");
			}

			public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html)).append(" Reflexive");
			}

			public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}
				b.append(keyword("Characteristics", html)).append(
						" InverseFunctional");
			}

			public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
				b.append(keyword("EquivalentTo", html)).append(" ");
				boolean first = true;
				for (final OWLObjectPropertyExpression d : arg0.getProperties()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI));
						first = false;
					}
				}

			}

			public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
				b.append(keyword("DisjointWith", html)).append(" ");
				boolean first = true;
				for (final OWLObjectPropertyExpression d : arg0.getProperties()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI));
						first = false;
					}
				}
			}

			public void visit(OWLObjectPropertyDomainAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}

				b.append(keyword("Domain", html)).append(" ").append(
						write(arg0.getDomain(), fullURI, html));
			}

			public void visit(OWLObjectPropertyRangeAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}

				b.append(keyword("Range", html)).append(" ").append(
						write(arg0.getRange(), fullURI, html));
			}

			public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
				// TODO
				boolean first = true;
				for (final OWLObjectPropertyExpression e : arg0
						.getPropertyChain()) {
					if (!first) {
						b.append(" ").append(keyword("o", html)).append(" ");
					}
					first = false;
					b.append(write(e, fullURI));
				}

				b.append(keyword("SubPropertyOf", html)).append("  ").append(
						write(arg0.getSuperProperty(), fullURI));
			}

			public void visit(OWLInverseObjectPropertiesAxiom arg0) {
				b.append(keyword("inverseOf", html)).append(" ");
				boolean first = true;
				for (final OWLObjectPropertyExpression d : arg0.getProperties()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI));
						first = false;
					}
				}

			}

			public void visit(OWLObjectSubPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getSubProperty(), fullURI)).append(" ");
				}

				b.append(keyword("SubPropertyOf", html)).append(" ").append(
						write(arg0.getSuperProperty(), fullURI));
			}

			public void visit(OWLDataPropertyDomainAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}

				b.append(keyword("Domain", html)).append(" ").append(
						write(arg0.getDomain(), fullURI, html));
			}

			public void visit(OWLDifferentIndividualsAxiom arg0) {
				b.append(keyword("DifferentFrom", html)).append(" ");
				boolean first = true;
				for (final OWLIndividual d : arg0.getIndividuals()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI, html));
						first = false;
					}
				}
			}

			public void visit(OWLDisjointDataPropertiesAxiom arg0) {
				b.append(keyword("DisjointWith", html)).append(" ");
				boolean first = true;
				for (final OWLDataPropertyExpression d : arg0.getProperties()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI));
						first = false;
					}
				}
			}

			public void visit(OWLDisjointUnionAxiom arg0) {
				// TODO
				b.append(write(arg0.getOWLClass(), fullURI, html)).append("=");
				b.append(keyword("DisjointUnionOf", html)).append("(");

				boolean first = true;

				for (final OWLDescription d : arg0.getDescriptions()) {
					if (!first) {
						b.append(", ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append(")");
			}

			public void visit(OWLDataPropertyRangeAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}

				b.append(keyword("Range", html)).append(" ").append(
						write(arg0.getRange()));
			}

			public void visit(OWLFunctionalDataPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getProperty(), fullURI)).append(" ");
				}

				b.append(keyword("Characteristics", html))
						.append(" Functional");
			}

			public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
				b.append(keyword("EquivalentTo", html)).append(" ");
				boolean first = true;
				for (final OWLDataPropertyExpression d : arg0.getProperties()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI));
						first = false;
					}
				}
			}

			public void visit(OWLNegativeObjectPropertyAssertionAxiom arg0) {

				if (context == null) {
					b.append(write(arg0.getSubject(), fullURI, html)).append(
							" ");
					b.append(keyword("not", html));
				} else {
					b.append(keyword("Fact: not", html)).append(" ");
				}

				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(write(arg0.getObject(), fullURI, html));
			}

			public void visit(OWLObjectPropertyAssertionAxiom arg0) {
				b.append(keyword("Fact:", html)).append(" ");

				if (context == null) {
					b.append(write(arg0.getSubject(), fullURI, html)).append(
							" ");
				}
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(write(arg0.getObject(), fullURI, html));
			}

			public void visit(OWLClassAssertionAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getIndividual(), fullURI, html))
							.append(" ");
				}

				b.append(keyword("Type:", html)).append(" ").append(
						write(arg0.getDescription(), fullURI, html));
			}

			public void visit(OWLDataPropertyAssertionAxiom arg0) {
				b.append(keyword("Fact:", html)).append(" ");

				if (context == null) {
					b.append(write(arg0.getSubject(), fullURI, html)).append(
							" ");
				}
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(write(arg0.getObject()));
			}

			public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
				b.append(keyword("Fact: not", html)).append(" ");
				if (context == null) {
					b.append(write(arg0.getSubject(), fullURI, html)).append(
							" ");
				}
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(write(arg0.getObject()));
			}

			public void visit(OWLDataSubPropertyAxiom arg0) {
				if (context == null) {
					b.append(write(arg0.getSubProperty(), fullURI)).append(" ");
				}
				b.append(keyword("SubPropertyOf", html)).append(" ");
				b.append(write(arg0.getSuperProperty(), fullURI));
			}

			public void visit(OWLSameIndividualsAxiom arg0) {
				b.append(keyword("SameAs", html)).append(" ");
				boolean first = true;
				for (final OWLIndividual d : arg0.getIndividuals()) {
					if (!d.equals(context)) {
						if (!first)
							b.append(", ");
						b.append(write(d, fullURI, html));
						first = false;
					}
				}
			}

			public void visit(SWRLRule arg0) {
				b.append(arg0.toString());
				// TODO
			}

			public void visit(OWLDeclarationAxiom arg0) {
				// b.append(arg0.toString());
				b.append(keyword("Declaration", html)).append(" ");
				b.append(getName(arg0.getEntity().getURI(), fullURI));
				// tohle jsem si vymyslel, protoze jsem to v Manchester syntaxi
				// nenasel
			}

			public void visit(OWLEntityAnnotationAxiom arg0) {
				if (context == null) {
					arg0.getSubject().accept(new OWLEntityVisitor() {

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLClass cls) {
							// TODO Auto-generated method stub
							b.append(write(cls, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLObjectProperty property) {
							b.append(write(property, fullURI));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLDataProperty property) {
							b.append(write(property, fullURI));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLIndividual individual) {
							b.append(write(individual, fullURI, html));
						}

						/**
						 * {@inheritDoc}
						 */
						public void visit(OWLDataType dataType) {
							b.append(write(dataType));
						}
					});

					b.append(" ");
				}

				b.append(keyword("Annotation", html)).append(" ");

				b.append(" ").append(
						getName(arg0.getAnnotation().getAnnotationURI(),
								fullURI));
				if (arg0.getAnnotation().isAnnotationByConstant()) {
					b.append(" ").append(
							write(arg0.getAnnotation()
									.getAnnotationValueAsConstant()));
				} else {
					// TODO
					b.append(" ").append(
							arg0.getAnnotation().getAnnotationValue()).append(
							"\"");
				}
			}

			public void visit(OWLOntologyAnnotationAxiom arg0) {
				b.append(keyword("Annotation", html)).append(" ");
				b.append(" ").append(
						getName(arg0.getAnnotation().getAnnotationURI(),
								fullURI));

				if (arg0.getAnnotation().isAnnotationByConstant()) {
					b.append(" ").append(
							write(arg0.getAnnotation()
									.getAnnotationValueAsConstant()));
				} else {
					// TODO
					b.append(" \"").append(
							arg0.getAnnotation().getAnnotationValue()).append(
							"\"");
				}
			}

			public void visit(OWLImportsDeclaration arg0) {
				b.append(keyword("Import", html)).append(" ");
				b.append(arg0.getImportedOntologyURI());
				// TODO
			}

			public void visit(OWLAxiomAnnotationAxiom arg0) {
				b.append(keyword("Annotation", html)).append(" ");
				b.append(getName(arg0.getAnnotation().getAnnotationURI(),
						fullURI));
				b.append(" \"").append(
						arg0.getAnnotation().getAnnotationValue()).append("\"");
				// TODO
			}

		});
		return b.toString();
	}

	public String write(final OWLDescription concept, final boolean fullURI,
			final boolean html) {
		if (concept == null) {
			throw new IllegalArgumentException();
		}

		final StringBuilder b = new StringBuilder();

		concept.accept(new OWLDescriptionVisitor() {

			public void visit(OWLClass arg0) {
				if (arg0.isOWLThing()) {
					b.append(keyword("Thing", html));
				} else if (arg0.isOWLThing()) {
					b.append(keyword("Nothing", html));
				} else {
					b.append(getName(arg0.getURI(), fullURI));
				}
			}

			public void visit(OWLObjectIntersectionOf arg0) {
				boolean first = true;
				boolean previous = false;

				for (final OWLDescription d : arg0.getOperands()) {
					if (!first) {
						b.append(" ");
						b.append(d.isAnonymous() && !previous ? keyword("that",
								html) : keyword("and", html));
						b.append(" ");
					}
					previous = d.isAnonymous();
					first = false;
					b.append(write(d, fullURI, html));

				}
			}

			public void visit(OWLObjectUnionOf arg0) {
				boolean first = true;

				for (final OWLDescription d : arg0.getOperands()) {
					if (!first) {
						b.append(" ");
						b.append(keyword("or", html));
						b.append(" ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}
			}

			public void visit(OWLObjectComplementOf arg0) {
				b.append(keyword("not", html)).append(" ");
				b.append(write(arg0.getOperand(), fullURI, html));
			}

			public void visit(OWLObjectSomeRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("some", html)).append(" ").append(
								write(arg0.getFiller(), fullURI, html));
			}

			public void visit(OWLObjectAllRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("only", html)).append(" ").append(
								write(arg0.getFiller(), fullURI, html));
			}

			public void visit(OWLObjectValueRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("value", html)).append(" ").append(
								write(arg0.getValue(), fullURI, html));
			}

			public void visit(OWLObjectMinCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("min", html)).append(" ").append(
								arg0.getCardinality());

				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
			}

			public void visit(OWLObjectExactCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("exactly", html)).append(" ").append(
								arg0.getCardinality());

				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
			}

			public void visit(OWLObjectMaxCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("max", html)).append(" ").append(
								arg0.getCardinality());

				if (!arg0.getFiller().isOWLThing()) {
					b.append(" ")
							.append(write(arg0.getFiller(), fullURI, html));
				}
			}

			public void visit(OWLObjectSelfRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ");
			}

			public void visit(OWLObjectOneOf arg0) {
				b.append("{");
				boolean first = true;

				for (final OWLIndividual d : arg0.getIndividuals()) {
					if (!first) {
						b.append(", ");
					}
					first = false;
					b.append(write(d, fullURI, html));
				}

				b.append("}");
			}

			public void visit(OWLDataSomeRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("some", html)).append(" ").append(
								write(arg0.getFiller()));
			}

			public void visit(OWLDataAllRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("only", html)).append(" ").append(
								write(arg0.getFiller()));
			}

			public void visit(OWLDataValueRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("value", html)).append(" ").append(
								write(arg0.getValue()));
			}

			public void visit(OWLDataMinCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("min", html)).append(" ").append(
								arg0.getCardinality());

				b.append(" ").append(write(arg0.getFiller()));
			}

			public void visit(OWLDataExactCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("exactly", html)).append(" ").append(
								arg0.getCardinality());

				b.append(" ").append(write(arg0.getFiller()));
			}

			public void visit(OWLDataMaxCardinalityRestriction arg0) {
				b.append(write(arg0.getProperty(), fullURI)).append(" ")
						.append(keyword("max", html)).append(" ").append(
								arg0.getCardinality());
				b.append(" ").append(write(arg0.getFiller()));
			}
		});

		return b.toString();
	}

	public String write(final OWLConstant c) {
		if (c.isTyped()) {
			final OWLTypedConstant tc = c.asOWLTypedConstant();
			return "\"" + tc.getLiteral() + "\"^^" + write(tc.getDataType());
		} else {
			final OWLUntypedConstant tc = c.asOWLUntypedConstant();
			return "\"" + tc.getLiteral() + "\"";
		}
	}

	public String write(final OWLDataRange dt) {
		return dt.toString();
	}

	public String write(final OWLPropertyExpression<?, ?> relation,
			final boolean fullURI) {
		if (relation == null) {
			throw new IllegalArgumentException("Relation is null.");
		}

		final StringBuilder b = new StringBuilder();

		relation.accept(new OWLPropertyExpressionVisitor() {

			public void visit(OWLObjectProperty arg0) {
				b.append(getName(arg0.getURI(), fullURI));
			}

			public void visit(OWLObjectPropertyInverse arg0) {
				b.append(write(arg0.getInverse(), fullURI));
			}

			public void visit(OWLDataProperty arg0) {
				b.append(getName(arg0.getURI(), fullURI));
			}

		});

		return b.toString();
	}

	private static String getName(final URI uri, final boolean fullURI) {
		if (fullURI) {
			return uri.toString();
		} else if (uri.getFragment() == null) {
			return uri.getPath().substring(uri.getPath().lastIndexOf('/') + 1);
		} else {

			return uri.getFragment();
		}
	}

	public String write(OWLIndividual individual, boolean fullURI, boolean html) {
		final StringBuffer b = new StringBuffer();

		if (html) {
			b.append("<i>");
		}

		b.append(getName(individual.getURI(), fullURI));

		if (html) {
			b.append("</i>");
		}

		return b.toString();
	}
}
