/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.cex;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAxiom;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectSomeRestriction;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLSubClassAxiom;

import cz.cvut.kbss.owldiff.OWLEntailmentChecker;

public class DiffL {

	private static final Log LOG = LogFactory.getLog(DiffL.class);

	private OWLOntology original, update;
	private Sig origSig, updateSig, sigma;
	private Noimply nio, niu;

	private Set<OWLClassPair> oo;
	private Map<OWLClass, Set<OWLClass>> oos;
	private Set<OWLClass> leftOver = null;

	public DiffL(OWLOntology original, OWLOntology update, Sig origSig,
			Sig updateSig, Sig sigma, Noimply nio, Noimply niu) {
		this.original = original;
		this.update = update;
		this.origSig = origSig;
		this.updateSig = updateSig;
		this.sigma = sigma;
		this.nio = nio;
		this.niu = niu;
	}

	public void generateOs() {
		generateO();
		oos = new HashMap<OWLClass, Set<OWLClass>>();

		Set<OWLClass> osig = new HashSet<OWLClass>();
		osig.addAll(origSig.getSig());
		Iterator<OWLClass> iter = osig.iterator();
		while (iter.hasNext()) {
			OWLClass a = iter.next();
			if (niu.primitive(niu.getAxioms(a))) {
				oos.put(a, getOClassesRightTo(a));
				iter.remove();
				continue;
			}

			Set<OWLClassAxiom> axioms = niu.getAxioms(a);
			if (axioms.size() != 1) {
				continue;
			}
			OWLClassAxiom axiom = axioms.iterator().next();
			OWLDescription d;
			// TODO bud je chyba v paperu, nebo to "a" (v paperu A') ma byt v
			// axiomech na prave strane
			if (axiom instanceof OWLEquivalentClassesAxiom) {
				OWLEquivalentClassesAxiom equAxiom = (OWLEquivalentClassesAxiom) axiom;
				if (equAxiom.getDescriptions().size() != 2) {
					continue;
				}
				d = Noimply.otherDesription(equAxiom.getDescriptions(), a);
			} else if (axiom instanceof OWLSubClassAxiom) {
				OWLSubClassAxiom subAxiom = (OWLSubClassAxiom) axiom;
				d = subAxiom.getSuperClass();
			} else {
				continue;
			}
			if (!(d instanceof OWLObjectSomeRestriction)) {
				continue;
			}
			OWLObjectSomeRestriction restr = (OWLObjectSomeRestriction) d;
			// if (!(restr.getFiller() instanceof OWLClass)) {continue;}
			if (!(restr.getProperty() instanceof OWLObjectProperty)) {
				continue;
			}
			// OWLClass a1p = (OWLClass) restr.getFiller();
			OWLObjectProperty r = (OWLObjectProperty) restr.getProperty();

			if (!sigma.getRoles().contains(r)) {
				oos.put(a, getOClassesRightTo(a));
				iter.remove();
			}

		}

		iter = osig.iterator();
		while (iter.hasNext()) {
			OWLClass a = iter.next();
			Set<OWLClassAxiom> axioms = niu.getAxioms(a);
			if (axioms.size() != 1) {
				continue;
			}
			OWLClassAxiom axiom = axioms.iterator().next();
			if (!(axiom instanceof OWLEquivalentClassesAxiom)) {
				continue;
			}
			OWLEquivalentClassesAxiom equAxiom = (OWLEquivalentClassesAxiom) axiom;
			if (equAxiom.getDescriptions().size() != 2) {
				continue;
			}
			OWLDescription od = Noimply.otherDesription(equAxiom
					.getDescriptions(), a);
			if (!(od instanceof OWLObjectIntersectionOf)) {
				continue;
			}
			OWLObjectIntersectionOf inters = (OWLObjectIntersectionOf) od;

			Set<OWLClass> o = getOClassesRightTo(a);
			for (OWLDescription d : inters.getOperands()) {
				// TODO pouzit radeji visitor ???
				if (!(d instanceof OWLClass)) {
					continue;
				}

				Set<OWLClass> ooax = oos.get(d);
				if (ooax != null) {
					o.retainAll(ooax);
				} else {
					o.clear();
					break;
				}
			}

			oos.put(a, o);
			iter.remove();
		}

		if (osig.size() == 0) {
			return;
		}
		OWLEntailmentChecker em = nio.getEM();
		OWLDataFactory fact = nio.getFactory();

		iter = osig.iterator();
		while (iter.hasNext()) {
			OWLClass ap = iter.next();
			Set<OWLClassAxiom> axioms = niu.getAxioms(ap);
			if (axioms.size() != 1) {
				continue;
			}
			OWLClassAxiom axiom = axioms.iterator().next();
			OWLDescription d;
			if (axiom instanceof OWLEquivalentClassesAxiom) {
				OWLEquivalentClassesAxiom equAxiom = (OWLEquivalentClassesAxiom) axiom;
				if (equAxiom.getDescriptions().size() != 2) {
					continue;
				}
				d = Noimply.otherDesription(equAxiom.getDescriptions(), ap);
			} else if (axiom instanceof OWLSubClassAxiom) {
				OWLSubClassAxiom subAxiom = (OWLSubClassAxiom) axiom;
				d = subAxiom.getSuperClass();
			} else {
				continue;
			}
			if (!(d instanceof OWLObjectSomeRestriction)) {
				continue;
			}
			OWLObjectSomeRestriction restr = (OWLObjectSomeRestriction) d;
			if (!(restr.getFiller() instanceof OWLClass)) {
				continue;
			}
			if (!(restr.getProperty() instanceof OWLObjectProperty)) {
				continue;
			}
			OWLClass a1p = (OWLClass) restr.getFiller();
			OWLObjectProperty r = (OWLObjectProperty) restr.getProperty();
			if (!sigma.getRoles().contains(r)) {
				continue;
			}

			Set<OWLClass> o = getOClassesRightTo(ap);
			Iterator<OWLClass> iter2 = o.iterator();
			while (iter2.hasNext()) {
				OWLClass a = iter2.next();

				boolean exists = false;
				Set<OWLClass> a1poos = oos.get(a1p);
				if (a1poos != null)
					for (OWLClass a1 : a1poos) {
						OWLAxiom ax = fact.getOWLSubClassAxiom(a, fact
								.getOWLObjectSomeRestriction(r, a1));
						if (em.isEntailed(ax)) {
							exists = true;
							break;
						}
					}
				if (!exists) {
					iter2.remove();
				}
			}
			oos.put(ap, o);
			iter.remove();
		}

		if (osig.size() > 0) {
			LOG.warn("DiffL: left over classes: " + Diff.classesToString(osig));
			leftOver = osig;
		}

	}

	private Set<OWLClass> getOClassesRightTo(OWLClass a) {
		Set<OWLClass> set = new HashSet<OWLClass>();
		for (OWLClassPair p : oo) {
			if (a.equals(p.a1)) {
				set.add(p.a2);
			}
		}
		return set;
	}

	private void generateO() {
		oo = new HashSet<OWLClassPair>();
		Set<OWLClass> osig = origSig.getSig();
		Set<OWLClass> sig = sigma.getSig();
		OWLEntailmentChecker emo = nio.getEM();
		OWLEntailmentChecker emu = niu.getEM();
		OWLDataFactory facto = nio.getFactory();
		OWLDataFactory factu = niu.getFactory();

		for (OWLClass ap : osig) {
			for (OWLClass a : osig) {
				boolean ok = true;

				for (OWLClass b : sig) {

					OWLAxiom ax = factu.getOWLSubClassAxiom(ap, b);
					if (!emu.isEntailed(ax)) {
						continue;
					}

					ax = facto.getOWLSubClassAxiom(a, b);
					if (emo.isEntailed(ax)) {
						continue;
					}

					ok = false;
					break;
				}

				if (ok) {
					OWLClassPair p = new OWLClassPair();
					p.a1 = ap;
					p.a2 = a;
					oo.add(p);
				}
			}
		}
	}

	public Set<OWLClassPair> getO() {
		return oo;
	}

	public Set<OWLClass> diffL() {
		if (oos == null) {
			LOG.error("DiffL: not generated O(a) yet!");
			return null;
		}
		Set<OWLClass> res = new HashSet<OWLClass>();
		for (OWLClass a : sigma.getSig()) {
			Set<OWLClass> ooa = oos.get(a);
			if ((ooa == null) || !ooa.contains(a)) {
				res.add(a);
			}
		}
		// if (leftOver != null) {res.removeAll(leftOver);} // TODO - je to v
		// poradku ?
		return res;
	}

	public class OWLClassPair {
		OWLClass a1;
		OWLClass a2;

		public boolean equals(Object obj) {
			if (!(obj instanceof OWLClassPair)) {
				return false;
			}
			if (this == obj) {
				return true;
			}
			OWLClassPair p = (OWLClassPair) obj;
			if ((p.a1 == a1) && (p.a2 == a2)) {
				return true;
			}
			if ((a1 == null) && (p.a1 != null)) {
				return false;
			}
			if ((a2 == null) && (p.a2 != null)) {
				return false;
			}
			if ((a1 != null) && !a1.equals(p.a1)) {
				return false;
			}
			if ((a2 != null) && !a2.equals(p.a2)) {
				return false;
			}
			return true;
		}

		public int hashCode() {
			int h = 0;
			if (a1 != null) {
				h += a1.hashCode();
			}
			if (a2 != null) {
				h += a2.hashCode();
			}
			return h;
		}

		public String toString() {
			return "[" + a1 + "," + a2 + "]";
		}
	}

}
