/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.cex;

import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.model.OWLAntiSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLAxiomAnnotationAxiom;
import org.semanticweb.owl.model.OWLAxiomVisitor;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAssertionAxiom;
import org.semanticweb.owl.model.OWLClassAxiom;
import org.semanticweb.owl.model.OWLDataAllRestriction;
import org.semanticweb.owl.model.OWLDataExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLDataMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLDataPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLDataPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLDataSomeRestriction;
import org.semanticweb.owl.model.OWLDataSubPropertyAxiom;
import org.semanticweb.owl.model.OWLDataValueRestriction;
import org.semanticweb.owl.model.OWLDeclarationAxiom;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLDescriptionVisitor;
import org.semanticweb.owl.model.OWLDifferentIndividualsAxiom;
import org.semanticweb.owl.model.OWLDisjointClassesAxiom;
import org.semanticweb.owl.model.OWLDisjointDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointUnionAxiom;
import org.semanticweb.owl.model.OWLEntityAnnotationAxiom;
import org.semanticweb.owl.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owl.model.OWLEquivalentDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLEquivalentObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLFunctionalDataPropertyAxiom;
import org.semanticweb.owl.model.OWLFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLInverseFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLIrreflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLNegativeDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLNegativeObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectAllRestriction;
import org.semanticweb.owl.model.OWLObjectComplementOf;
import org.semanticweb.owl.model.OWLObjectExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectOneOf;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyChainSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLObjectSelfRestriction;
import org.semanticweb.owl.model.OWLObjectSomeRestriction;
import org.semanticweb.owl.model.OWLObjectSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectUnionOf;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyAnnotationAxiom;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLReflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLSameIndividualsAxiom;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.model.OWLSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTransitiveObjectPropertyAxiom;
import org.semanticweb.owl.model.SWRLRule;

import cz.cvut.kbss.owldiff.OWLEntailmentChecker;
import cz.cvut.kbss.owldiff.pellet.PelletEntailmentChecker;

public class Noimply {

	private static final Log LOG = LogFactory.getLog(Noimply.class);

	private OWLOntology ont;
	private Sig sig;
	private Sig sigma;
	private Map<OWLClass, Set<OWLClass>> pres;

	private OWLAxiom alpha;
	private OWLDescription allSigmaSuperClass;
	private OWLClass allSigmaClass;
	private Set<OWLDescription> sigmaFresh;
	private Map<OWLClass, OWLClass> xiClasses;
	private Map<OWLClass, OWLClass> xiClassesInv;
	private Map<OWLClass, OWLDescription> xiSuperClasses;
	private Map<OWLClass, OWLAxiom> alphas;
	private Map<OWLClass, Set<OWLClass>> noImplies;
	private Map<OWLClass, Set<OWLAxiom>> nnoImplies;
	private Set<OWLAxiom> nnoImply;
	private Map<OWLClass, Set<OWLClassAxiom>> classAxioms;
	private Set<OWLClass> xxi;

	private OWLDataFactory factory;
	private final OWLEntailmentChecker em;

	public Noimply(OWLOntologyManager m, OWLOntology ont, Sig sig,
			OWLDataFactory factory, Sig sigma) {
		this.ont = ont;
		this.sig = sig;
		this.factory = factory;
		this.sigma = sigma;
		em = getExplanationManager(m, ont);
		pres = new HashMap<OWLClass, Set<OWLClass>>();
		alphas = new HashMap<OWLClass, OWLAxiom>();
		noImplies = new HashMap<OWLClass, Set<OWLClass>>();
		nnoImplies = new HashMap<OWLClass, Set<OWLAxiom>>();
		classAxioms = new HashMap<OWLClass, Set<OWLClassAxiom>>();

		buildAllSigma();
		buildSigmaFresh();

		xxi = new HashSet<OWLClass>();
		for (OWLClass a : sig.getSig()) {
			if (!conjunctive(a)) {
				xxi.add(xi(a));
			}
		}
	}

	private void buildAllSigma() {
		Set<OWLClass> s2 = new HashSet<OWLClass>();
		s2.addAll(sigma.getSig());
		try {
			allSigmaClass = factory.getOWLClass(new URI("#allSigma"));
		} catch (Exception ex) {
			LOG.error("Noimply allSigma creation exception", ex);
			return;
		}
		s2.add(allSigmaClass);
		OWLObjectIntersectionOf inters = factory.getOWLObjectIntersectionOf(s2);
		Set<OWLDescription> interset = new HashSet<OWLDescription>();
		for (OWLObjectProperty role : sigma.getRoles()) {
			OWLDescription d = factory
					.getOWLObjectSomeRestriction(role, inters);
			interset.add(d);
		}
		allSigmaSuperClass = factory.getOWLObjectIntersectionOf(interset);
		alpha = factory.getOWLSubClassAxiom(allSigmaClass, allSigmaSuperClass);
	}

	private void buildSigmaFresh() {
		sigmaFresh = new HashSet<OWLDescription>();
		sigmaFresh.add(allSigmaClass);
		xiClasses = new HashMap<OWLClass, OWLClass>();
		xiClassesInv = new HashMap<OWLClass, OWLClass>();
		xiSuperClasses = new HashMap<OWLClass, OWLDescription>();
		for (OWLClass a : sig.getSig()) {
			if (!conjunctive(a)) {
				sigmaFresh.add(xi(a));
			}
		}
	}

	private boolean conjunctive(OWLClass a) {
		Set<OWLClassAxiom> axioms = getAxioms(a);
		return conjunctive(a, axioms);
	}

	private boolean conjunctive(OWLClass a, Set<OWLClassAxiom> axioms) {

		if (pseudoPrimitive(axioms)) {
			return false;
		}

		boolean nonconj = true;
		for (OWLClassAxiom axiom : axioms) {
			if (!(axiom instanceof OWLEquivalentClassesAxiom)) {
				nonconj = false;
				break;
			} else {
				OWLEquivalentClassesAxiom equiAxiom = (OWLEquivalentClassesAxiom) axiom;
				Set<OWLDescription> set = new HashSet<OWLDescription>();
				set.addAll(equiAxiom.getDescriptions());
				set.remove(a);
				if (set.size() != 1) {
					nonconj = false;
					break;
				} else {
					OWLDescription d = set.iterator().next();
					if (!(d instanceof OWLObjectSomeRestriction)) {
						nonconj = false;
						break;
					}
				}
			}
		}
		return !nonconj;
	}

	public boolean primitive(Set<OWLClassAxiom> axioms) {
		return (axioms.size() == 0);
	}

	public boolean pseudoPrimitive(Set<OWLClassAxiom> axioms) {
		if (primitive(axioms)) {
			return true;
		}

		for (OWLClassAxiom axiom : axioms) {
			if (!(axiom instanceof OWLSubClassAxiom)) {
				return false;
			}
		}
		return true;
	}

	public Set<OWLClassAxiom> getAxioms(OWLClass a) {
		Set<OWLClassAxiom> axioms = classAxioms.get(a);
		if (axioms != null) {
			return axioms;
		}

		axioms = ont.getAxioms(a);
		classAxioms.put(a, axioms);
		return axioms;
	}

	public Set<OWLClass> noImply(OWLClass a) {
		Set<OWLClass> r = noImplies.get(a);
		if (r != null) {
			return r;
		}

		Set<OWLClassAxiom> axioms = getAxioms(a);
		r = new HashSet<OWLClass>();

		if (pseudoPrimitive(axioms)) {
			r.add(xi(a));
		} else if (axioms.size() == 1) {
			OWLClassAxiom axiom = axioms.iterator().next();
			axiom.accept(new NoImplyForSingleAxiomVisitor(a, r, conjunctive(a,
					axioms)));
		} else {
			LOG.warn("Noimply: noImply: Too many axioms for a class");
		}

		noImplies.put(a, r);
		return r;
	}

	public Set<OWLAxiom> nnoImply(OWLClass a) {
		Set<OWLAxiom> r = nnoImplies.get(a);
		if (r != null) {
			return r;
		}

		Set<OWLClassAxiom> axioms = getAxioms(a);
		r = new HashSet<OWLAxiom>();

		if (pseudoPrimitive(axioms)) {
			r.add(factory.getOWLSubClassAxiom(xi(a), xiSuperClass(a)));
		} else if (axioms.size() == 1) {
			OWLClassAxiom axiom = axioms.iterator().next();
			axiom.accept(new NNoImplyForSingleAxiomVisitor(a, r, conjunctive(a,
					axioms)));
		} else {
			LOG.warn("Noimply: nnoImply: Too many axioms for a class");
		}

		nnoImplies.put(a, r);
		return r;
	}

	public static OWLDescription otherDesription(Set<OWLDescription> ds,
			OWLDescription other) {
		for (OWLDescription d : ds) {
			if (d.equals(other)) {
				continue;
			}
			return d;
		}
		return null;
	}

	public Set<OWLAxiom> nnoImply() {
		if (nnoImply != null) {
			return nnoImply;
		}
		Set<OWLClass> s = sigma.getSig();
		s.addAll(sig.getSig());
		Set<OWLAxiom> set = new HashSet<OWLAxiom>();
		set.add(alpha);
		for (OWLClass c : s) {
			set.addAll(nnoImply(c));
		}
		nnoImply = set;
		return set;
	}

	/*
	 * private void addAllConjuncts(Set<OWLClass> set, OWLObjectIntersectionOf
	 * inter) { for (OWLDescription d : inter.getOperands()) { if (d instanceof
	 * OWLClass) { set.add((OWLClass) d); } else if (d instanceof
	 * OWLObjectIntersectionOf) { addAllConjuncts(set, (OWLObjectIntersectionOf)
	 * d); } else {LOG.warn(
	 * "Unsupported intersection axiom part in (supposed to be) normalized terminology"
	 * ); } } }
	 */

	public Set<OWLClass> pre(OWLClass a) {
		Set<OWLClass> set = pres.get(a);
		if (set != null) {
			return set;
		}
		OWLClass a2 = xiinv(a);
		if (a2 != null) {
			if (!conjunctive(a2)) {
				return pres.get(a2);
			} else {
				return null;
			}
		}
		set = new HashSet<OWLClass>();
		for (OWLClass b : sigma.getSig()) {
			OWLAxiom ax = factory.getOWLSubClassAxiom(b, a);
			if (em.isEntailed(ax)) {
				set.add(b);
			}
		}
		pres.put(a, set);
		return set;
	}

	private OWLClass xi(OWLClass a) {
		OWLClass xi = xiClasses.get(a);
		if (xi != null) {
			return xi;
		}

		try {
			xi = factory
					.getOWLClass(new URI("#xi_" + a.getURI().getFragment()));
		} catch (Exception ex) {
			LOG.error("Noimply xi class creation exception: " + ex);
			return null;
		}

		xiClasses.put(a, xi);
		xiClassesInv.put(xi, a);
		return xi;
	}

	public OWLClass xiinv(OWLClass xi) {
		return xiClassesInv.get(xi);
	}

	private OWLDescription xiSuperClass(OWLClass a) {
		OWLDescription xsc = xiSuperClasses.get(a);
		if (xsc != null) {
			return xsc;
		}

		HashSet<OWLClass> set = new HashSet<OWLClass>();
		set.add(allSigmaClass);
		set.addAll(sigma.getSig());
		set.removeAll(pre(a));
		xsc = factory.getOWLObjectIntersectionOf(set);

		xiSuperClasses.put(a, xsc);
		return xsc;
	}

	private OWLAxiom alpha(OWLClass a, OWLObjectProperty r, OWLClass b) {
		OWLAxiom alpha = alphas.get(a);
		if (alpha != null) {
			return alpha;
		}

		Set<OWLDescription> interset = new HashSet<OWLDescription>();
		interset.addAll(sigma.getSig());
		interset.removeAll(pre(a));

		for (OWLObjectProperty s : sigma.getRoles()) {
			if (s.equals(r)) {
				continue;
			}
			Set<OWLDescription> interset2 = new HashSet<OWLDescription>();
			interset2.addAll(sigma.getSig());
			interset2.add(allSigmaClass);
			interset.add(factory.getOWLObjectSomeRestriction(s, factory
					.getOWLObjectIntersectionOf(interset2)));
		}

		for (OWLClass xi : noImply(b)) {
			interset.add(factory.getOWLObjectSomeRestriction(r, xi));
		}

		OWLObjectIntersectionOf inters = factory
				.getOWLObjectIntersectionOf(interset);
		alpha = factory.getOWLSubClassAxiom(xi(a), inters);

		alphas.put(a, alpha);
		return alpha;
	}

	/*
	 * public class Result { Set<OWLClass> n = new HashSet<OWLClass>();
	 * Set<OWLAxiom> nn = new HashSet<OWLAxiom>(); }
	 */

	private OWLEntailmentChecker getExplanationManager(
			final OWLOntologyManager m, final OWLOntology o) {
		return new PelletEntailmentChecker(m, o);
	}

	public Set<OWLClass> getXi() {
		return xxi;
	}

	public OWLDataFactory getFactory() {
		return factory;
	}

	public OWLClass getAllSigma() {
		return allSigmaClass;
	}

	public OWLEntailmentChecker getEM() {
		return em;
	}

	public OWLAxiom getAlpha() {
		return alpha;
	}

	private class NoImplyForSingleAxiomVisitor implements OWLAxiomVisitor,
			OWLDescriptionVisitor {
		private OWLClass a;
		private Set<OWLClass> r;
		private boolean conjunctive;
		private boolean equivalentAxiom = false;

		public NoImplyForSingleAxiomVisitor(OWLClass a, Set<OWLClass> r,
				boolean conjunctive) {
			this.a = a;
			this.r = r;
			this.conjunctive = conjunctive;
		}

		private void unusedAxiom(OWLAxiom axiom) {
			LOG.warn("Unused axiom: " + axiom);
		}

		private void unusedDescription(OWLDescription descr) {
			LOG.warn("Unused description: " + descr);
		}

		public void visit(OWLSubClassAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLNegativeObjectPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointClassesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataPropertyDomainAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLImportsDeclaration arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLAxiomAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyDomainAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDifferentIndividualsAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointDataPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyRangeAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointUnionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDeclarationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEntityAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLOntologyAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataPropertyRangeAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLFunctionalDataPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLClassAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentClassesAxiom axiom) {
			Set<OWLDescription> descs = axiom.getDescriptions();
			if (descs.size() != 2) {
				return;
			}
			OWLDescription d = otherDesription(descs, a);
			if (d == null) {
				return;
			}
			equivalentAxiom = true;
			d.accept(this);
		}

		public void visit(OWLDataPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLSameIndividualsAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLInverseObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(SWRLRule arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLClass cls) {
			if (conjunctive && equivalentAxiom) {
				r.add(xi(cls));
			} else {
				unusedDescription(cls);
				// r.n.add(a);
				// TODO r.nn.add();
			}
		}

		public void visit(OWLObjectIntersectionOf inters) {
			if (conjunctive && equivalentAxiom) {
				for (OWLDescription d : inters.getOperands()) {
					d.accept(this);
				}
			} else {
				unusedDescription(inters);
				// addAllConjuncts(r.n, inters);
				// r.nn = 0
			}
		}

		public void visit(OWLObjectUnionOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectComplementOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectSomeRestriction rest) {
			if (conjunctive && equivalentAxiom) {
				rest.getFiller().accept(this);
			} else {
				if ((rest.getProperty() instanceof OWLObjectProperty)
						&& (rest.getFiller() instanceof OWLClass)) {
					r.add(xi(a));
				} else {
					LOG
							.warn("Noimply visitor: unexpected params of OWLObjectSomeRestriction");
				}
			}

		}

		public void visit(OWLObjectAllRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectValueRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectSelfRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectOneOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataSomeRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataAllRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataValueRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}
	}

	private class NNoImplyForSingleAxiomVisitor implements OWLAxiomVisitor,
			OWLDescriptionVisitor {
		private OWLClass a;
		private Set<OWLAxiom> r;
		private boolean conjunctive;
		private boolean equivalentAxiom = false;

		public NNoImplyForSingleAxiomVisitor(OWLClass a, Set<OWLAxiom> r,
				boolean conjunctive) {
			this.a = a;
			this.r = r;
			this.conjunctive = conjunctive;
		}

		private void unusedAxiom(OWLAxiom axiom) {
			LOG.warn("Unused axiom: " + axiom);
		}

		private void unusedDescription(OWLDescription descr) {
			LOG.warn("Unused description: " + descr);
		}

		public void visit(OWLSubClassAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLNegativeObjectPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointClassesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataPropertyDomainAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLImportsDeclaration arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLAxiomAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyDomainAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDifferentIndividualsAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointDataPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyRangeAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDisjointUnionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDeclarationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEntityAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLOntologyAnnotationAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataPropertyRangeAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLFunctionalDataPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLClassAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLEquivalentClassesAxiom axiom) {
			Set<OWLDescription> descs = axiom.getDescriptions();
			if (descs.size() != 2) {
				return;
			}
			OWLDescription d = otherDesription(descs, a);
			if (d == null) {
				return;
			}
			equivalentAxiom = true;
			d.accept(this);
		}

		public void visit(OWLDataPropertyAssertionAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLDataSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLSameIndividualsAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLInverseObjectPropertiesAxiom arg0) {
			unusedAxiom(arg0);
		}

		public void visit(SWRLRule arg0) {
			unusedAxiom(arg0);
		}

		public void visit(OWLClass cls) {
			if (conjunctive && equivalentAxiom) {

			} else {
				unusedDescription(cls);
				// r.n.add(a);
				// TODO r.nn.add();
			}
		}

		public void visit(OWLObjectIntersectionOf inters) {
			if (conjunctive && equivalentAxiom) {
				/*
				 * for (OWLDescription d : inters.getOperands()) {
				 * d.accept(this); }
				 */
			} else {
				LOG
						.warn("Nopimply visitor: unexpected OWLObjectIntersectionOf");
				// addAllConjuncts(r.n, inters);
				// r.nn = 0
			}
		}

		public void visit(OWLObjectUnionOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectComplementOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectSomeRestriction rest) {
			// TODO Auto-generated method stub
			if (conjunctive && equivalentAxiom) {
				// rest.getFiller().accept(this);
			} else {
				if ((rest.getProperty() instanceof OWLObjectProperty)
						&& (rest.getFiller() instanceof OWLClass)) {
					r.add(alpha(a, (OWLObjectProperty) rest.getProperty(),
							(OWLClass) rest.getFiller()));
				} else {
					LOG
							.warn("Nopimply visitor: unexpected params of OWLObjectSomeRestriction");
				}
			}

		}

		public void visit(OWLObjectAllRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectValueRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectSelfRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLObjectOneOf arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataSomeRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataAllRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataValueRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		public void visit(OWLDataMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}
	}

}
