/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.cex;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.AddAxiom;
import org.semanticweb.owl.model.OWLAntiSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLAxiomAnnotationAxiom;
import org.semanticweb.owl.model.OWLAxiomVisitor;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAssertionAxiom;
import org.semanticweb.owl.model.OWLDataAllRestriction;
import org.semanticweb.owl.model.OWLDataExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLDataMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLDataPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLDataPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLDataSomeRestriction;
import org.semanticweb.owl.model.OWLDataSubPropertyAxiom;
import org.semanticweb.owl.model.OWLDataValueRestriction;
import org.semanticweb.owl.model.OWLDeclarationAxiom;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLDescriptionVisitor;
import org.semanticweb.owl.model.OWLDifferentIndividualsAxiom;
import org.semanticweb.owl.model.OWLDisjointClassesAxiom;
import org.semanticweb.owl.model.OWLDisjointDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLDisjointUnionAxiom;
import org.semanticweb.owl.model.OWLEntityAnnotationAxiom;
import org.semanticweb.owl.model.OWLEquivalentClassesAxiom;
import org.semanticweb.owl.model.OWLEquivalentDataPropertiesAxiom;
import org.semanticweb.owl.model.OWLEquivalentObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLFunctionalDataPropertyAxiom;
import org.semanticweb.owl.model.OWLFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLImportsDeclaration;
import org.semanticweb.owl.model.OWLInverseFunctionalObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLIrreflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLNegativeDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLNegativeObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectAllRestriction;
import org.semanticweb.owl.model.OWLObjectComplementOf;
import org.semanticweb.owl.model.OWLObjectExactCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectMaxCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectMinCardinalityRestriction;
import org.semanticweb.owl.model.OWLObjectOneOf;
import org.semanticweb.owl.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyChainSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyDomainAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyRangeAxiom;
import org.semanticweb.owl.model.OWLObjectSelfRestriction;
import org.semanticweb.owl.model.OWLObjectSomeRestriction;
import org.semanticweb.owl.model.OWLObjectSubPropertyAxiom;
import org.semanticweb.owl.model.OWLObjectUnionOf;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyAnnotationAxiom;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLReflexiveObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLSameIndividualsAxiom;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.model.OWLSymmetricObjectPropertyAxiom;
import org.semanticweb.owl.model.OWLTransitiveObjectPropertyAxiom;
import org.semanticweb.owl.model.SWRLRule;

import cz.cvut.kbss.owldiff.OWLDiffException;
import cz.cvut.kbss.owldiff.ProgressListener;
import cz.cvut.kbss.owldiff.OWLDiffException.Reason;

public class Normalizator implements OWLAxiomVisitor {

	private static final Log LOG = LogFactory.getLog(Normalizator.class);

	private static final int PROGRESS_MAX = 100;

	OWLOntologyManager manager;

	OWLDataFactory factory;

	private OWLOntology ont;
	private OWLOntology nont;

	private int errorCnt;

	private ProgressListener progressListener;

	public Normalizator(OWLOntology ont, ProgressListener progressListener) {
		this.ont = ont;
		this.progressListener = progressListener;
		try {
			manager = OWLManager.createOWLOntologyManager();
			factory = manager.getOWLDataFactory();
			nont = manager.createOntology(ont.getURI());
		} catch (Exception ex) {
			LOG.error("Failed to init normalizator", ex);
		}
	}

	public OWLOntology normalizeTerminology() throws OWLDiffException {
		int i = 0;
		if (progressListener != null) {
			progressListener.setProgress(0);
			progressListener.setProgressMax(PROGRESS_MAX);
		}

		errorCnt = 0;
		for (OWLAxiom axiom : ont.getAxioms()) {

			axiom.accept(this);

			if (progressListener != null) {
				progressListener.setProgress((++i) % PROGRESS_MAX);
			}

		}

		if (errorCnt > 0) {
			throw new OWLDiffException(Reason.INCOMPATIBLE_ONTOLOGY,
					"Ontology " + ont.getURI() + " is not EL");
		}

		return nont;
	}

	List<OWLDescription> getListOfConjunctions(OWLDescription in) {
		List<OWLDescription> list = new ArrayList<OWLDescription>();
		new NormalizatorDescriptionVisitor(in, list);
		return list;
	}

	List<OWLObjectIntersectionOf> getConjunctionsFromLists(
			List<List<OWLDescription>> lists) {
		List<OWLObjectIntersectionOf> conjs = new ArrayList<OWLObjectIntersectionOf>();
		for (List<OWLDescription> list : lists) {
			conjs.add(factory
					.getOWLObjectIntersectionOf(new HashSet<OWLDescription>(
							list)));
		}
		return conjs;
	}

	@SuppressWarnings("unchecked")
	private List<List<OWLDescription>> getCartesianProduct(
			List<List<OWLDescription>> lists) {
		List<List<OWLDescription>> out = new ArrayList<List<OWLDescription>>();
		if (lists.size() == 0) {
			return out;
		}
		if (lists.size() == 1) {
			for (OWLDescription d : lists.get(0)) {
				List<OWLDescription> l = new ArrayList<OWLDescription>();
				l.add(d);
				out.add(l);
			}
			return out;
		}
		List<OWLDescription> ll = lists.remove(lists.size() - 1);
		List<List<OWLDescription>> fout = getCartesianProduct(lists);
		for (List<OWLDescription> flist : fout) {
			for (OWLDescription d : ll) {
				List<OWLDescription> l = (List<OWLDescription>) ((ArrayList<OWLDescription>) flist)
						.clone();
				l.add(d);
				out.add(l);
			}
		}
		return out;
	}

	private void addAxiom(OWLOntology ont, OWLAxiom axiom) {
		try {
			manager.applyChange(new AddAxiom(ont, axiom));
		} catch (Exception ex) {
			LOG.error("Failed to add axiom " + axiom + " to ontology " + ont);
		}
	}

	private void unusedAxiom(OWLAxiom axiom) {
		LOG.warn("Ignored axiom: " + axiom);
		errorCnt++;
	}

	
	public void visit(OWLSubClassAxiom axiom) {
		OWLDescription sub = axiom.getSubClass();
		OWLDescription sup = axiom.getSuperClass();
		if (sub.isOWLNothing()) {
			return;
		}
		if (sup.isOWLThing()) {
			return;
		}
		if (!(sub instanceof OWLClass)) {
			LOG
					.error("On the left side of a subAxiom is not a single class, skipping");
			errorCnt++;
			return;
		}
		// OWLClass subClass = (OWLClass) sub;
		List<OWLDescription> list = getListOfConjunctions(sup);
		for (OWLDescription d : list) {
			OWLAxiom a = factory.getOWLSubClassAxiom(sub, d);
			addAxiom(nont, a);
		}
	}

	
	public void visit(OWLNegativeObjectPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDisjointClassesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDataPropertyDomainAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLImportsDeclaration arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLAxiomAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLObjectPropertyDomainAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDifferentIndividualsAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDisjointDataPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLObjectPropertyRangeAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLObjectPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLObjectSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDisjointUnionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDeclarationAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLEntityAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLOntologyAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDataPropertyRangeAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLFunctionalDataPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLClassAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLEquivalentClassesAxiom axiom) {
		java.util.Set<OWLDescription> descrs = axiom.getDescriptions();
		OWLClass left = null;
		OWLDescription right = null;
		if (descrs.size() != 2) {
			LOG.error("Unsupported equAxiom - not 2 sides");
			errorCnt++;
			return;
		}
		for (OWLDescription d : descrs) {
			if ((d instanceof OWLClass) && (left == null)) {
				left = (OWLClass) d;
			} else {
				right = d;
			}
		}
		if (left == null) {
			LOG
					.error("Unsupported equAxiom - at least one side has to be class");
			errorCnt++;
			return;
		}
		List<OWLDescription> list = getListOfConjunctions(right);
		// TODO does it make sense to break down equivalence axiom like this ???
		// Probably not:
		if (list.size() != 1) {
			LOG.error("Unsupported equAxiom - union of some things");
			errorCnt++;
			return;
		}

		for (OWLDescription d : list) {
			Set<OWLDescription> set = new HashSet<OWLDescription>();
			set.add(left);
			set.add(d);
			OWLAxiom a = factory.getOWLEquivalentClassesAxiom(set);
			addAxiom(nont, a);
		}

	}

	
	public void visit(OWLDataPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLDataSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLSameIndividualsAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(OWLInverseObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	
	public void visit(SWRLRule arg0) {
		unusedAxiom(arg0);
	}

	private class NormalizatorDescriptionVisitor implements
			OWLDescriptionVisitor {

		private List<OWLDescription> list;

		public NormalizatorDescriptionVisitor(OWLDescription in,
				List<OWLDescription> list) {
			this.list = list;

			in.accept(this);

			/*
			 * } else { LOG.error("Unsupported OWLDescription type"); }
			 */
		}

		private void unusedDescription(OWLDescription descr) {
			LOG.warn("Ignored description: " + descr);
			errorCnt++;
		}

		
		public void visit(OWLClass c) {
			list.add(c);
		}

		
		public void visit(OWLObjectIntersectionOf inters) {

			Set<OWLDescription> ops = inters.getOperands();
			List<List<OWLDescription>> lists = new ArrayList<List<OWLDescription>>();
			for (OWLDescription op : ops) {
				lists.add(getListOfConjunctions(op));
			}
			list.addAll(getConjunctionsFromLists(getCartesianProduct(lists)));

		}

		
		public void visit(OWLObjectUnionOf union) {

			Set<OWLDescription> ops = union.getOperands();
			for (OWLDescription op : ops) {
				/*
				 * ArrayList<OWLDescription> l = new
				 * ArrayList<OWLDescription>(); new
				 * NormalizatorDescriptionVisitor(op, l); list.addAll(l);
				 */
				list.addAll(getListOfConjunctions(op));
			}

		}

		
		public void visit(OWLObjectComplementOf arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectSomeRestriction restr) {

			list.add(restr);

		}

		
		public void visit(OWLObjectAllRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectValueRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectSelfRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLObjectOneOf arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataSomeRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataAllRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataValueRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataMinCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataExactCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

		
		public void visit(OWLDataMaxCardinalityRestriction arg0) {
			unusedDescription(arg0);
		}

	}

}
