/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.cex;

import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owl.model.*;

public class Sig implements OWLAxiomVisitor,OWLDescriptionVisitor,OWLPropertyExpressionVisitor {
	
	private static final Log LOG = LogFactory.getLog(Sig.class);

	private HashSet<OWLClass> sig;
	private HashSet<OWLObjectProperty> roles;
	
	private Sig() {
		sig = new HashSet<OWLClass>();
		roles = new HashSet<OWLObjectProperty>();
	}
	
	public Sig(OWLOntology o) {
		this();
		for (OWLAxiom axiom : o.getAxioms()) {
			axiom.accept(this);
		}
		
	}

	public Sig(Set<OWLClass> classes, Set<OWLObjectProperty> properties) {
		this();
		if (classes != null) {sig.addAll(classes);}
		if (properties != null) {roles.addAll(properties);}
	}

	@SuppressWarnings("unchecked")
	public Set<OWLClass> getSig() {
		return (Set<OWLClass>) sig.clone();
	}
	
	@SuppressWarnings("unchecked")
	public Set<OWLObjectProperty> getRoles() {
		return (Set<OWLObjectProperty>) roles.clone();
	}
	
	public void removeClass(OWLClass c) {
		sig.remove(c);
	}
	
	public void removeRole(OWLObjectProperty r) {
		roles.remove(r);
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("Sig: classes: ");
		for (OWLClass c : sig) {sb.append(c.getURI().getFragment()+", ");}
		sb.append("; roles: ");
		for (OWLObjectProperty p : roles) {sb.append(p.getURI().getFragment()+", ");}
		return sb.toString();
	}

	
	private void unusedAxiom(OWLAxiom axiom) {
		LOG.debug("Unused axiom: "+axiom);
	}
	private void unusedDescription(OWLDescription descr) {
		LOG.debug("Unused description: "+descr);
	}
	private void unusedProperty(OWLPropertyExpression prop) {
		LOG.debug("Unused property: "+prop);
	}

	
	
	public void visit(OWLSubClassAxiom axiom) {
		axiom.getSubClass().accept(this);
		axiom.getSuperClass().accept(this);
	}

	public void visit(OWLNegativeObjectPropertyAssertionAxiom axiom) {
		axiom.getProperty().accept(this);
		
	}

	public void visit(OWLAntiSymmetricObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLReflexiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDisjointClassesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDataPropertyDomainAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLImportsDeclaration arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLAxiomAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLObjectPropertyDomainAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLEquivalentObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLNegativeDataPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDifferentIndividualsAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDisjointDataPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDisjointObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLObjectPropertyRangeAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLObjectPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLFunctionalObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLObjectSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDisjointUnionAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDeclarationAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLEntityAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLOntologyAnnotationAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLSymmetricObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDataPropertyRangeAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLFunctionalDataPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLEquivalentDataPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLClassAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLEquivalentClassesAxiom axiom) {
		for (OWLDescription d : axiom.getDescriptions()) {
			d.accept(this);
		}		
	}

	public void visit(OWLDataPropertyAssertionAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLTransitiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLIrreflexiveObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLDataSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLInverseFunctionalObjectPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLSameIndividualsAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLObjectPropertyChainSubPropertyAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(OWLInverseObjectPropertiesAxiom arg0) {
		unusedAxiom(arg0);
	}

	public void visit(SWRLRule arg0) {
		unusedAxiom(arg0);
	}


	
	
	public void visit(OWLClass cls) {
		sig.add(cls); 
	}

	public void visit(OWLObjectIntersectionOf inters) {
		for (OWLDescription d : inters.getOperands()) {
			d.accept(this);
		}
	}

	public void visit(OWLObjectUnionOf arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectComplementOf arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectSomeRestriction restr) {
		restr.getFiller().accept(this);
		restr.getProperty().accept(this);
	}

	public void visit(OWLObjectAllRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectValueRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectMinCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectExactCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectMaxCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectSelfRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLObjectOneOf arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataSomeRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataAllRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataValueRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataMinCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataExactCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	public void visit(OWLDataMaxCardinalityRestriction arg0) {
		unusedDescription(arg0);
	}

	
	
	
	public void visit(OWLObjectProperty prop) {
		roles.add(prop);
	}

	
	public void visit(OWLObjectPropertyInverse inv) {
		inv.getInverse().accept(this);
		
	}

	
	public void visit(OWLDataProperty prop) {
		unusedProperty(prop);
	}
	
	
}
