/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.pellet;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mindswap.pellet.owlapi.Reasoner;
import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.AddAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyChangeException;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;

import cz.cvut.kbss.explanation.Test;
import cz.cvut.kbss.owldiff.OWLEntailmentChecker;

public class PelletEntailmentTest implements Test<OWLAxiom> {

	private static final Log LOG = LogFactory
			.getLog(PelletEntailmentTest.class);

	private int i;

	private OWLAxiom toTest;

	private enum State {
		TAUTOLOGY, UNKNOWN;
	}

	private State s = State.UNKNOWN;

	public void setAxiom(final OWLAxiom a) {
		toTest = a;
		s = State.UNKNOWN;
		try {
			final OWLOntologyManager m = OWLManager.createOWLOntologyManager();
			final OWLOntology o = m.createOntology(URI.create("" + i++));

			final List<AddAxiom> adds = new ArrayList<AddAxiom>();

			// add referenced named entities to avoid Pellet crash
			for (final OWLEntity e : toTest.getReferencedEntities()) {
				adds.add(new AddAxiom(o, m.getOWLDataFactory()
						.getOWLDeclarationAxiom(e)));
			}

			m.applyChanges(adds);

			final OWLEntailmentChecker r = new PelletEntailmentChecker(m, o);

			if (r.isEntailed(toTest)) {
				s = State.TAUTOLOGY;
			}
		} catch (OWLOntologyCreationException e) {
			LOG.error(e, e);
		} catch (OWLOntologyChangeException e) {
			LOG.error(e, e);
		}

	}

	public boolean test(Set<OWLAxiom> v) {
		if (s.equals(State.TAUTOLOGY)) {
			return true; // no explanations for tautologies !
		}

		final OWLOntologyManager m = OWLManager.createOWLOntologyManager();

		final Reasoner r = new Reasoner(m);

		try {
			final OWLOntology o = m.createOntology(URI.create("" + i++));

			final List<AddAxiom> adds = new ArrayList<AddAxiom>();

			// add referenced named entities to avoid Pellet crash
			for (final OWLEntity e : toTest.getReferencedEntities()) {
				adds.add(new AddAxiom(o, m.getOWLDataFactory()
						.getOWLDeclarationAxiom(e)));
			}

			for (final OWLAxiom a : v) {
				adds.add(new AddAxiom(o, a));
			}

			m.applyChanges(adds);
			r.loadOntology(o);
			return !r.isEntailed(toTest);
		} catch (OWLOntologyCreationException e) {
			LOG.error(e, e);
		} catch (OWLOntologyChangeException e) {
			LOG.error(e, e);
		}

		return false;
	}
}
