/*******************************************************************************
 * Copyright (c) 2008 Czech Technical University in Prague.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v2.1
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 ******************************************************************************/
package cz.cvut.kbss.owldiff.pellet;

import java.util.List;

import org.mindswap.pellet.owlapi.PelletVisitor;
import org.mindswap.pellet.owlapi.Reasoner;
import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLEntityVisitor;

import cz.cvut.kbss.explanation.IncrementalTest;
import cz.cvut.kbss.explanation.IncrementalTestState;

class PelletIncrementalTest implements IncrementalTest<OWLAxiom> {

	private OWLAxiom toTest;

	private Reasoner createEmptyState() {
		final Reasoner r = new Reasoner(OWLManager.createOWLOntologyManager());

		// add referenced named entities to avoid Pellet crash
		for (final OWLEntity e : toTest.getReferencedEntities()) {
			e.accept((OWLEntityVisitor) new PelletVisitor(r.getKB()));
		}

		return r;
	}

	public void setAxiom(final OWLAxiom a) {
		this.toTest = a;
	}

	public Object copyState(Object object) {
		if (object == null) {
			return createEmptyState();
		} else {
			Reasoner rOld = (Reasoner) object;
			// // FIXME - proper copy of KB
			return new Reasoner(OWLManager.createOWLOntologyManager(), rOld
					.getKB().copy());
		}
	}

	/**
	 * WARNING : no proper copy of the KB is performed.
	 * 
	 * {@inheritDoc}
	 */
	public IncrementalTestState test(OWLAxiom v, Object state) {
		Reasoner s = (state == null) ? createEmptyState() : (Reasoner) state;

		final PelletVisitor visitor = new PelletVisitor(s.getKB());
		visitor.setAddAxiom(true);
		v.accept(visitor);

		return new IncrementalTestState(!s.isEntailed(toTest), s);
	}

	public IncrementalTestState test(List<OWLAxiom> component, Object state) {
		Reasoner s = (state == null) ? createEmptyState() : (Reasoner) state;

		final PelletVisitor visitor = new PelletVisitor(s.getKB());
		visitor.setAddAxiom(true);

		for (final OWLAxiom a : component) {
			a.accept(visitor);
		}

		return new IncrementalTestState(!s.isEntailed(toTest), s);
	}
}
